;;;-*- Mode: Lisp; Package: (PPC :use CL ARCH) -*-

;;	Change History (most recent first):
;;  28 9/12/96 slh  
;;  10 12/2/95 gb   add %INIT-MISC nrs
;;  5 10/26/95 gb   %gcable-pointers% not a symbol; %macro-code% is.  Go figure.  Some other globals, kernel-imports.
;;  4 10/23/95 gb   added %gcable-macptrs% to nrs
;;  2 10/6/95  gb   Minor changes.
;;  (do not edit before this line!!)

; Modification History
;
; 03/13/97 bill  error-fpu-exception-single from Gary.
; 02/14/97 gb    fp_s32conv
; 02/13/97 bill  Add to *ppc-nilreg-relative-symbols* :
;                ccl::*cs-segment-size*, ccl::*vs-segment-size*,
;                ccl::*ts-segment-size*, ccl::*cs-hard-overflow-size*,
;                ccl::*cs-soft-overflow-size*, ccl::cs-overflow-callback
; 10/22/96 bill  error-too-many-values
; ------------   4.0
; 09/02/96 gb    vsp comment was lying: stacks grow towards 0.
;  8/28/96 slh   nrs *current-cursor*
; --- 4.0b1
; 07/25/96 gb    kernel-import lisp-bug.
; 07/08/96 bill  protected-area storage layout
; 06/14/96 bill  Add %new-gcable-ptr to *ppc-nilreg-relative-symbols*
; 06/20/96 gb    fp-zero (= fp31).  area.nextref-> area.threshold, area.gc-count.
;                kernel-globals oldest-ephemeral, tenured-area.  Import egc-control.
; --- 3.9 release ---
; 04/10/96 gb    error-memory-full.
; 04/08/96 gb    define subtag-go-tag & subtag-block-tag; nrs %excised-code%,
;                error-excised-function-call, kernel-import excise-library.
; 04/08/96 bill  add exception-saved-registers to *ppc-kernel-globals*
; 03/25/96 bill  error-divide-by-0 => error-throw-tag-missing to agree
;                with "ccl:pmcl;errors.s".
; 03/01/96 bill  More cross-reference comment for *ppc-kernel-globals*
;                Add resolve-slep-address to *ppc-nilreg-relative-symbols*
; 03/10/96 gb    new globals & import; creole; fp error, fpscr bits
; 03/01/96 gb    new AREA fields
; 01/22/96 bill  add ndwords to (define-storage-layout area ...)
;                Add comment about gc-area record definition.
; 02/01/96 gb    kernel-import-condemn-area.  Area stuff (missing ndwords!)
;                lexpr-return, lexpr-return1v.
; 01/22/96 gb    error-stack-overflow, error-alloc-failed.
; 12/27/95 gb    remove pad words from ratio, complex; add gc nilreg globals.
; 12/13/95 gb    add %builtin-functions% nrs, builtin-function-name-offset
; 12/09/95 bill  add to kernel imports: allocate_tstack, allocate_vstack, register_cstack
; 12/09/95 bill  gb: add nilreg-globals current-cs, current-vs,
;                current-ts, cs-overflow-limit, all-areas.
; 11/30/95 bill  comments on error numbers
; 11/10/95 gb    new type-error numbering.
; 10/30/95 bill  Add some cross-references to "ccl:pmcl;"
; 10/25/95 slh   updated *ppc-kernel-globals*

;; This file matches "ccl:pmcl;constants.h" & "ccl:pmcl;constants.s"

(eval-when (:compile-toplevel :load-toplevel :execute)
  (require "ARCH"))

(defpackage "PPC" (:use "CL" "ARCH"))

(in-package "PPC")



;; PPC-32 stuff and tags.
(eval-when (:compile-toplevel :load-toplevel :execute)


;;; Lisp registers.
(eval-when (:compile-toplevel :execute)
  (defmacro defregs (&body regs)
    `(progn
       (defenum () ,@regs)
       (defparameter *gpr-register-names* ,(coerce (mapcar #'string regs) 'vector))))
  (defmacro deffpregs (&body regs)
    `(progn
       (defenum () ,@regs)
       (defparameter *fpr-register-names* ,(coerce (mapcar #'string regs) 'vector)))))

(defregs
  rzero                                 ; Always contains 0; not as handy as it sounds.
  sp                                    ; The control stack.  Aligned on 16-byte boundary.
  rnil                                  ; Keep NIL here; don't use TOC.
  imm0                                  ; Unboxed, volatile registers.
  imm1 
  imm2 
  imm3 
  imm4
  nargs                                 ; Volatile.  SHOULDN'T be used for tag extraction. (TWI handler confusion.)
  freeptr                               ; Heap free pointer. Shared among lisp threads.
  initptr                               ; Heap initialization pointer.  Can't take an interrupt while consing.
  memo                                  ; EGC memoization stack pointer.
  tsp                                   ; Temp-stack pointer.
  vsp                                   ; Value stack pointer; grows towards 0.
  loc-g                                 ; Node locative
  loc-pc                                ; For return PC only.
  fn                                    ; Current function (constants vector).
  temp3                                 ; Boxed, volatile registers.  Some may be defined on function entry.
  temp2 
  temp1 
  temp0 
  arg_x                                 ; Next-to-next-to-last function argument.
  arg_y                                 ; Next-to-last function argument.
  arg_z                                 ; Last function argument.
  save7                                 ; Boxed, nonvolatile registers.
  save6
  save5
  save4 
  save3 
  save2 
  save1 
  save0
)

(deffpregs 
  fp0
  fp1
  fp2
  fp3
  fp4
  fp5
  fp6
  fp7
  fp8
  fp9
  fp10
  fp11
  fp12
  fp13
  fp14
  fp15
  fp16
  fp17
  fp18
  fp19
  fp20
  fp21
  fp22
  fp23
  fp24
  fp25
  fp26
  fp27
  fp28
  fp29
  fp30
  fp31)


(defconstant fname temp3)

;;; Calling sequence may pass additional arguments in temp registers.
;;; "nfn" (new function) is always passed; it's the new value of "fn".
(defconstant nfn temp2)

;;; CLOS may pass the context for, e.g.., CALL-NEXT-METHOD in 
;;;; the "next-method-context" register.
(defconstant next-method-context temp1)


;;; It's handy to have 0.0 in an fpr.
(defconstant fp-zero fp31)

; Also handy to have #x4330000080000000 in an fpr, for s32->float conversion.
(defconstant fp-s32conv fp30)




(defconstant max-64-bit-constant-index (ash (+ #x7fff misc-dfloat-offset) -3))
(defconstant max-32-bit-constant-index (ash (+ #x7fff misc-data-offset) -2))
(defconstant max-16-bit-constant-index (ash (+ #x7fff misc-data-offset) -1))
(defconstant max-8-bit-constant-index (+ #x7fff misc-data-offset))
(defconstant max-1-bit-constant-index (ash (+ #x7fff misc-data-offset) 5))


; The objects themselves look something like this:


;; For the eabi port: mark this stack frame as Lisp's (since EABI
;; foriegn frames can be the same size as a lisp frame.)


(define-storage-layout lisp-frame 0
  backlink
  savefn
  savelr
  savevsp
)

(define-storage-layout c-frame 0
  backlink
  crsave
  savelr
  unused-1
  unused-2
  savetoc
  param0
  param1
  param2
  param3
  param4
  param5
  param6
  param7
)

(defconstant c-frame.minsize c-frame.size)

;; .SPeabi-ff-call "shrinks" this frame after loading the GPRs.
(define-storage-layout eabi-c-frame 0
  backlink
  savelr
  param0
  param1
  param2
  param3
  param4
  param5
  param6
  param7
)

(defconstant eabi-c-frame.minsize eabi-c-frame.size)






(defenum (:prefix "PPC-" :suffix "-BIT")
  lt
  gt
  eq
  so
)


(defenum (:prefix "FPSCR-" :suffix "-BIT")
  fx
  fex
  vx
  ox
  ux
  zx
  xx
  vxsnan
  vxisi
  vxidi
  vxzdz
  vximz
  vxvc
  fr
  fi
  fprfc
  fl
  fg
  fe
  fu
  nil
  vxsoft
  vxsqrt
  vxcvi
  ve
  oe
  ue
  ze
  xe
  ni
  rn0
  rn1
)

)
(ccl::provide "PPC-ARCH")
