;;; -*- Mode: Lisp; Package: CCL -*-


;;; Fix a code vector so that its subprim references are relative to (this image's) subprims-base.
;;; Then, do an FF-CALL to MakeDataExecutable so that the data cache gets flushed.
;;; This needs to be called inside a WITHOUT-INTERRUPTS.
(defppclapfunction %make-code-executable ((codev arg_z))
  (let ((len temp2))
    (let ((word-offset imm0)
          (unmasked-word imm1)
          (masked-word imm2)
          (ba-mask imm3)
          (ba-val imm4)
          (sp-base temp0)           ; tagged as a fixnum
          (i temp1))
      (save-lisp-context)
      (li i '0)
      (getvheader word-offset codev)
      (header-length len word-offset)
      (lwi ba-mask #.(dpb #x3f (byte 6 26) (ash 1 1)))
      (lwi ba-val  #.(dpb #x12 (byte 6 26) (ash 1 1)))
      (ref-global sp-base subprims-base)
      (li word-offset arch::misc-data-offset)
      @nextword
      (addi i i '1)
      (cmpw cr0 i len)
      (lwzx unmasked-word codev word-offset)
      (cmpwi cr2 unmasked-word 0)       ; traceback table follows ?
      (and masked-word unmasked-word ba-mask)
      (cmpw cr1 masked-word ba-val)
      (bne+ cr1 @nofix)
      (add unmasked-word unmasked-word sp-base)
      (stwx unmasked-word codev word-offset)
      @nofix
      (addi word-offset word-offset 4)
      (beq cr2 @done)                   ; don't relocate traceback table ...
      (bne cr0 @nextword))
    @done
    ;;; This is going to have to pass a lisp object to a foreign function.
    ;;; If we ever have a preemptive scheduler, we'd better hope that
    ;;; WITHOUT-INTERRUPTS refuses to run lisp code from a callback.
    (progn
      (stwu sp (- (+ ppc::eabi-c-frame.minsize ppc::lisp-frame.size)) sp)      ; make an FFI frame.
      (la imm0 arch::misc-data-offset codev)
      (stw imm0 ppc::eabi-c-frame.param0 sp)
      (stw len ppc::eabi-c-frame.param1 sp)
      (ref-global imm3 kernel-imports)
      (lwz arg_z arch::kernel-import-MakeDataExecutable imm3)
      (bla .SPeabi-ff-call))
    (mr arg_z rnil)
    (restore-full-lisp-context)
    (blr)))

; Except for a brief moment (before the compiler/fasloader has called
; %make-code-executable), a heap-consed code-vector has had the runtime
; value of subprims-base added to any of its subprim call instructions.
; A code-vector that's been copied to a read-only section makes PC-relative
; calls to a copy of the subprims library that sits at the beginning of
; that section.
; Some things (the fasdumper & disassembler, at least) would find it easier
; to deal with a copy of the code-vector in which all subprim calls are in
; a "normalized" (absolute 16-bit target address) form.  Normalizing a "pure"
; code vector means replacing pc-relative branches with absolute ones; normalizing
; a relocatable code-vectors means adjusting the absolute branch instructions.
; In both cases, "dest" should be some vector which contains at least as many
; bytes as are in the code vector.  Some things (the fasdumper for instance) may
; find it more efficient to re-use the same vector.

; This assumes that any pc-relative branch whose target is before the code-vector's
; entrypoint is a branch to a subprim.

(defun normalize-code-vector (code-vector 
                              &optional (dest 
                                         (let* ((size (uvsize code-vector))
                                                                  (v (make-array (the fixnum size)
                                                                                 :element-type '(unsigned-byte 32))))
                                                             (declare (fixnum size))
                                           (%copy-ivector-to-ivector code-vector 0 v 0 (the fixnum (ash size 2))))))
  (let* ((readonly-prefix (%readonly-area-word-offset code-vector 2)))
    (if readonly-prefix
      (%normalize-pure-code-vector readonly-prefix code-vector dest)
      (%normalize-heap-code-vector code-vector dest))))

(defppclapfunction %normalize-pure-code-vector ((prefix-words arg_x) (codev arg_y) (dest arg_z))
  (let ((len temp2))
    (let ((word-offset imm0)
          (unmasked-word imm1)
          (masked-word imm2)
          (brel-mask imm3)
          (brel-val imm4)
          (disp temp0)
          (i temp1)
          (curpc temp3)
          (target imm2))
      (li i '0)
      (mr curpc prefix-words)
      (getvheader word-offset codev)
      (header-length len word-offset)
      (lwi brel-mask #.(dpb #x3f (byte 6 26) (dpb 1 (byte 1 25) (ash 1 1))))    ; opcode, LI sign, BA bit
      (lwi brel-val  #.(dpb #x12 (byte 6 26) (dpb 1 (byte 1 25) (ash 0 1))))    ; negative relative branch
      (li word-offset arch::misc-data-offset)
      @nextword
      (addi i i '1)
      (cmpw cr0 i len)
      (lwzx unmasked-word codev word-offset)
      (cmpwi cr2 unmasked-word 0)       ; traceback table follows ?
      (and masked-word unmasked-word brel-mask)
      (cmpw cr1 masked-word brel-val)
      (bne+ cr1 @nofix)
      (extlwi disp unmasked-word 24 6)
      (srawi disp disp 6)
      (add target curpc disp)
      (cmpw cr1 target prefix-words)
      (rlwimi unmasked-word target 0 6 29)
      (ori unmasked-word unmasked-word (ash 1 1))       ; turn on AA bit
      (bge+ cr1 @nofix)
      (stwx unmasked-word dest word-offset)
      @nofix
      (addi word-offset word-offset 4)
      (addi curpc curpc '1)
      (beq cr2 @done)                   ; don't relocate traceback table ...
      (bne cr0 @nextword))
    @done
    (blr)))

(defppclapfunction %normalize-heap-code-vector ((codev arg_y) (dest arg_z))
  (let ((len temp2))
    (let ((word-offset imm0)
          (unmasked-word imm1)
          (masked-word imm2)
          (ba-mask imm3)
          (ba-val imm4)
          (sp-base temp0)           ; tagged as a fixnum
          (i temp1))
      (li i '0)
      (getvheader word-offset codev)
      (header-length len word-offset)
      (lwi ba-mask #.(dpb #x3f (byte 6 26) (ash 1 1)))
      (lwi ba-val  #.(dpb #x12 (byte 6 26) (ash 1 1)))
      (ref-global sp-base subprims-base)
      (li word-offset arch::misc-data-offset)
      @nextword
      (addi i i '1)
      (cmpw cr0 i len)
      (lwzx unmasked-word codev word-offset)
      (cmpwi cr2 unmasked-word 0)       ; traceback table follows ?
      (and masked-word unmasked-word ba-mask)
      (cmpw cr1 masked-word ba-val)
      (bne+ cr1 @nofix)
      (sub unmasked-word unmasked-word sp-base)
      (stwx unmasked-word dest word-offset)
      @nofix
      (addi word-offset word-offset 4)
      (beq cr2 @done)                   ; don't relocate traceback table ...
      (bne cr0 @nextword))
    @done
  (blr)))



    


; end of ppc-def.lisp
