;;;-*- Mode: Lisp; Package: CCL -*-

;;	Change History (most recent first):
;;  2 2/13/97  akh  #+ppc-target for byte-length
;;  9 9/4/96   akh  targetize architecture in get-shared-library?
;;  2 11/9/95  bill Temporarily nuke close-connection
;;  1 11/9/95  bill New file
;;  (do not edit before this line!!)


; l0-cfm-support.lisp
; Copyright 1995 Digitool, Inc. The 'tool rules!
; Support for the Code Fragment Manager

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Modification History
;;;
;;; 02/06/97 akh   define byte-length iff undefined
;;; 01/17/97 bill  Add "AppleGuideGlueLib" to *lazy-shared-libraries*
;;; 12/26/96 bill  Make close-connection work on the 68K.
;;; -------------  4.0
;;; 06/03/96 bill  Remove some obsolete comments.
;;; -------------  MCL-PPC 3.9
;;; 03/22/96 bill  reincarnate close-connection in a form that works in level 0.
;;;                reload-shared-library
;;; 03/22/96 bill  %kernel-import & %get-unboxed-ptr no longer store unboxed data in arg_z
;;; 03/11/96 bill  improve get-shared-library's error message.
;;; 03/08 96 bill  initialize-shared-libraries unconditionally calls clear-shared-library-caches
;;;                (I'm paranoid).
;;; 03/06/96 bill  Add more libraries to *lazy-shared-libraries*
;;; 03/05/96 bill  add "ObjectSupportLib", "QuickTimeLib", "DragLib", "TelephoneLib",
;;;                "ProcessMgrSupport", "Translation", "ColorPickerLib", "SpeechLib"
;;;                to *lazy-shared-libraries*
;;;                get-shared-library no longer conses the error-name return value if
;;;                it is the same as the library-name parameter.
;;; 03/01/96 bill  resolve-slep-address passes library-name to get-frag-symbol which uses
;;;                it to generate a better error message.
;;; 02/18/96 bill  *lazy-shared-libraries*, map-over-fixed-and-lazy-shared-libraries
;;;                This gives us a way to add automatically searched shared libraries
;;;                without always opening them.
;;; 01/17/96 bill  shared-library-search-path-entry, add-to-shared-library-search-path,
;;;                remove-from-shared-library-search-path
;;;


; offset is a fixnum, one of the arch::kernel-import-xxx.
; Returns that kernel import, a fixnum.
#+ppc-target
(defppclapfunction %kernel-import ((offset arg_z))
  (ref-global imm0 kernel-imports)
  (unbox-fixnum imm1 arg_z)
  (lwzx arg_z imm0 imm1)
  (blr))

#+sparc-target
(defsparclapfunction %kernel-import ((%offset %arg_z))
  (ref-global %imm0 kernel-imports)
  (unbox-fixnum %offset %imm1)
  (retl)
    (ld (%imm0 %imm1) %arg_z))

#+ppc-target
(defppclapfunction %get-unboxed-ptr ((macptr arg_z))
  (macptr-ptr imm0 arg_z)
  (lwz arg_z 0 imm0)
  (blr))

#+sparc-target
(defsparclapfunction %get-unboxed-ptr ((%macptr %arg_z))
  (macptr-ptr %macptr %imm0)
  (retl)
   (ld (%imm0) %arg_z))


; Bootstrapping. Real version is in l1-aprims.
; Called by expansion of with-pstrs

(defun byte-length (string &optional script start end)
    (declare (ignore script))
    (when (or start end)
      (error "Don't support start or end args yet"))
    (if (base-string-p string)
      (length string)
      (error "Don't support non base-string yet.")))



(def-accessor-macros %svref
  nil                                 ; 'external-entry-point
  eep.address
  eep.name
  eep.container)

(defun %cons-external-entry-point (name &optional container)
  (%istruct 'external-entry-point nil name container))

(defun external-entry-point-p (x)
  (istruct-typep x 'external-entry-point))

(defvar *rtld-next*)
(setq *rtld-next* (%int-to-ptr #xFFFFFFFF))


(defvar *eeps* nil)

(defun eeps ()
  (or *eeps*
      (setq *eeps* (make-hash-table :test #'equal))))

(defun resolve-container (c)
  (if c
    (error "Don't handle external containers yet.")
    *rtld-next*))

(defun resolve-eep (e)
  (or (eep.address e)
      (let* ((name (eep.name e))
             (handle (resolve-container (eep.container e))))
        (with-cstrs ((n name))
          (let* ((addr (ff-call (%kernel-import arch::kernel-import-dlsym) :address handle :address n :signed-fullword)))
            (declare (integer addr))
            (if (zerop addr)
              (error "Can't resolve external-entrypoint ~s" name)
              (setf (eep.address e) (ash addr -2))))))))

          
(defun load-eep (name)
  (let* ((eep (or (gethash name (eeps)) (setf (gethash name *eeps*) (%cons-external-entry-point name)))))
    (resolve-eep eep)
    eep))

#+ppc-target
(defppclapfunction %revive-macptr ((p arg_z))
  (li imm0 arch:subtag-macptr)
  (stb imm0 arch::misc-subtag-offset p)
  (blr))

#+sparc-target
(defsparclapfunction %revive-macptr ((p %arg_z))
  (mov arch::subtag-macptr %imm0)
  (retl)
  (stb %imm0 (p arch::misc-subtag-offset)))

(defun refresh-external-entrypoints ()
  (%revive-macptr *rtld-next*)
  (when *eeps*
    (without-interrupts 
     (maphash #'(lambda (k v) 
                  (declare (ignore k)) 
                  (setf (eep.address v) nil) 
                  (resolve-eep v))
              *eeps*))))




