; -*- Mode: Lisp; Package: CCL; -*-

(defconstant $seek-set 0)
(defconstant $seek-cur 1)
(defconstant $seek-end 2)

; write nbytes bytes from buffer buf to file-descriptor fd.
(defun fd-write (fd buf nbytes)
  (ff-call  
   (%kernel-import arch::kernel-import-write)
   :unsigned-fullword fd :address buf :unsigned-fullword nbytes :signed-fullword))

(defun fd-read (fd buf nbytes)
  (ff-call 
   (%kernel-import arch::kernel-import-read)
   :unsigned-fullword fd :address buf :unsigned-fullword nbytes :signed-fullword))

(defun fd-open (path flags &optional (create-mode #o666))
  (with-cstrs ((p path))
    (let* ((fd
            (ff-call 
             (%kernel-import arch::kernel-import-open)
             :address p :unsigned-fullword flags :unsigned-fullword create-mode :signed-fullword)))
      (declare (fixnum fd))
      (values fd (if (< fd 0) (%get-errno))))))

(defun fd-lseek (fd offset whence)
  (let* ((offset (ff-call 
                  (%kernel-import arch::kernel-import-lseek)
                  :unsigned-fullword fd
                  :unsigned-fullword offset
                  :unsigned-fullword whence
                  :signed-fullword)))
    (values offset (if (< offset 0) (%get-errno)))))


(defun fd-close (fd)
  (ff-call 
   (%kernel-import arch::kernel-import-close)
   :unsigned-fullword fd :signed-fullword))

(defun fd-tell (fd)
  (fd-lseek fd 0 $seek-cur))

(defun fd-size (fd)
  (let* ((curpos (fd-tell fd)))
    (unwind-protect
      (fd-lseek fd 0 $seek-end)
      (fd-lseek fd curpos $seek-set))))

(defun fd-ftruncate (fd new)
  (ff-call 
   (%kernel-import arch::kernel-import-ftruncate)
   :unsigned-fullword fd :unsigned-fullword new :signed-fullword))



;; Not really I/O, but ...
(defun malloc (size)
  (ff-call 
   (%kernel-import arch::kernel-import-malloc)
   :unsigned-fullword size :address))

(defun free (ptr)
  (ff-call 
   (%kernel-import arch::kernel-import-free)
   :address ptr :void))
