;;;-*- Mode: Lisp; Package: CCL -*-

;; l1-boot-2.lisp
;; Second part of l1-boot

;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Modification History
;;
;; 04/16/97 bill  New file. Broken out of l1-boot.lisp
;;


(macrolet ((l1-load (name)
	     (let* ((namestring
		     (concatenate 'simple-base-string
				  #+ppc-target "./l1-pfsls/"
				  #+sparc-target "./l1-sfsls/"
				  (string name)
				  #+ppc-target ".pfsl"
				  #+sparc-target ".sfsl")))
	       `(%fasload ,namestring)))
	   (bin-load (name)
	     (let* ((namestring
		     (concatenate 'simple-base-string
				  #+ppc-target "./binppc/"
				  #+sparc-target "./binsparc/"
				  (string name)
				  #+ppc-target ".pfsl"
				  #+sparc-target ".sfsl")))
	       `(%fasload ,namestring))))


(catch :toplevel
    (l1-load "l1-typesys")
    (l1-load "sysutils")
    #+ppc-target
    (l1-load "ppc-error-signal")
    #+sparc-target
    (l1-load "sparc-error-signal")
    (l1-load "l1-error-signal")
    (setq *LEVEL-1-LOADED* t))

(defvar *terminal-input* ())
(defvar *terminal-output* ())

;; fix this (I/O redirection!)
#+not-yet
(def-ccl-pointers fd-streams ()
  (let* ((ttyin (isatty 0))
         (ttyout (isatty 1))
         (ttyerr (isatty 2))
         (in (make-tty-stream 0 :direction :input :interactive ttyin))
         (out (make-tty-stream 1 :direction :output :interactive ttyout))
         (error (if (and ttyerr ttyout)
                    out
                    (make-tty-stream 2 :direction :output))))
    (setq *terminal-input* (if ttyin
                               in
                               (make-tty-stream
                                (fd-open "/dev/tty" O_RDONLY)))

          *terminal-output* (if ttyout
                                out
                                (make-tty-stream
                                 (fd-open "/dev/tty" O_WRONLY)
				 :direction :output)))
          
    (setq *terminal-io* (make-instance 'echoing-two-way-stream 
				       :input-stream *terminal-input*
				       :output-stream *terminal-output*))
    (setq *debug-io* (make-instance 'echoing-two-way-stream 
				    :input-stream *terminal-input*
				    :output-stream error)
          *query-io* *debug-io*)
    (setq *standard-input* (if ttyin *terminal-io* in)
          *standard-output* (if ttyout *terminal-io* out)
          *error-output* (if ttyerr *terminal-io* error)
          *trace-output* *error-output*)))

#-fixed-other-version-above
;; Assume that FD's 0,1, and 2 are all interactive streams; don't
;; try to open up /dev/tty (screws ilisp on Solaris.)

(def-ccl-pointers fd-streams ()
  (let* ((in (make-tty-stream 0 :direction :input :interactive t))
         (out (make-tty-stream 1 :direction :output :interactive t)))
    (setq *terminal-input* in
          *terminal-output* out)          
    (setq *terminal-io* (make-instance 'echoing-two-way-stream 
				       :input-stream *terminal-input*
				       :output-stream *terminal-output*))
    (setq *debug-io* *terminal-io*
          *query-io* *debug-io*)
    (setq *standard-input* in
          *standard-output* out
          *error-output* out
          *trace-output* *error-output*)))

(catch :toplevel
    (macrolet ((l1-load-provide (module path)
		 `(let* ((*package* *package*))
		   (l1-load ,path)
		   (provide ,module)))
	       (bin-load-provide (module path)
		 `(let* ((*package* *package*))
		   (bin-load ,path)
		   (provide ,module))))
      (bin-load-provide "SORT" "sort")
      (bin-load-provide "NUMBERS" "numbers")
      (bin-load-provide "HASH" "hash")
      (bin-load-provide "DLL-NODE" "dll-node")
      
      #+ppc-target
      (bin-load-provide "PPC-ARCH" "ppc-arch")
      #+sparc-target
      (bin-load-provide "SPARC-ARCH" "sparc-arch")
      
      (bin-load-provide "VREG" "vreg")
      
      #+ppc-target
      (bin-load-provide "PPC-ASM" "ppc-asm")
      #+sparc-target
      (bin-load-provide "SPARC-ASM" "sparc-asm")
      
      (bin-load-provide "VINSN" "vinsn")
      (bin-load-provide "REG" "reg")
      (bin-load-provide "SUBPRIMS" "subprims")
      
      #+ppc-target
      (bin-load-provide "PPC-LAP" "ppc-lap")
      #+sparc-target
      (bin-load-provide "SPARC-LAP" "sparc-lap")
      
      (bin-load-provide "BACKEND" "backend")
     
      #+ppc-target
      (provide "PPC2")                  ; Lie, load the module manually
      #+sparc-target
      (provide "SPARC2")
      
      (l1-load-provide "NX" "nx")
      
      #+ppc-target
      (bin-load "ppc2")
      #+sparc-target
      (bin-load "sparc2")
      
      (bin-load-provide "LEVEL-2" "level-2")
      (bin-load-provide "MACROS" "macros")
      (bin-load-provide "SETF" "setf")
      (bin-load-provide "SETF-RUNTIME" "setf-runtime")
      (bin-load-provide "FORMAT" "format")
      (bin-load-provide "STREAMS" "streams")
      (bin-load-provide "OPTIMIZERS" "optimizers")
      (progn  ; Shouldn't need this at load time ...
	#+ppc-target
	(%fasload "./library/lispequ.pfsl")
	#+sparc-target
	(%fasload "./library/lispequ.sfsl")
	(provide "LISPEQU"))      
      (bin-load-provide "DEFSTRUCT-MACROS" "defstruct-macros")        ;  ... but this file thinks it does.
      (bin-load-provide "DEFSTRUCT-LDS" "defstruct-lds")
      (bin-load-provide "NFCOMP" "nfcomp")
      (bin-load-provide "BACKQUOTE" "backquote")
      (bin-load-provide "BACKTRACE-LDS" "backtrace-lds")
      (bin-load-provide "BACKTRACE" "backtrace")
      (bin-load-provide "READ" "read")
      (bin-load-provide "ARRAYS-FRY" "arrays-fry")
      (bin-load-provide "APROPOS" "apropos")
      
      #+ppc-target
      (progn
	(bin-load-provide "PPC-DISASSEMBLE" "ppc-disassemble")
	(bin-load-provide "PPC-LAPMACROS" "ppc-lapmacros"))
      #+sparc-target
      (progn
	(bin-load-provide "SPARC-DISASSEMBLE" "sparc-disassemble")
	(bin-load-provide "SPARC-LAPMACROS" "sparc-lapmacros"))
      
      (bin-load-provide "MACTYPES" "mactypes")
      (bin-load-provide "DEFRECORD" "defrecord")
      #+ppc-target
      (progn
	(%fasload "./library/linux-records.pfsl")
	(provide "LINUX-RECORDS"))
      #+sparc-target
      (progn
	(%fasload "./library/solaris-records.sfsl")
	(provide "SOLARIS-RECORDS"))
      
      (bin-load-provide "CASE-ERROR" "case-error")
      (bin-load-provide "ENCAPSULATE" "encapsulate")
      (bin-load-provide "METHOD-COMBINATION" "method-combination")
      (bin-load-provide "MISC" "misc")
      (bin-load-provide "PPRINT" "pprint")
      (bin-load-provide "DUMPLISP" "dumplisp")
      (bin-load-provide "PATHNAMES" "pathnames")
      (bin-load-provide "TIME" "time")
      (bin-load-provide "COMPILE-CCL" "compile-ccl")
      (bin-load-provide "DESCRIBE" "describe")
      (bin-load-provide "SOURCE-FILES" "source-files")
      (bin-load-provide "CCL-EXPORT-SYMS" "ccl-export-syms")
      )
    (setq *%fasload-verbose* nil)
    )
)






