

;;	Change History (most recent first):
;;  1 11/13/95 gb   split off from l1-readloop.lisp
;;
;;  (do not edit before this line!!)
;;; Functions that actually try to signal or handle errors.

;; 03/26/97 gb    (defpascal %err-disp) : recognize errors during FRSP as "can't coerce".
;; -------------  4.1b1
;; 03/13/97 bill  From Gary: (defpascal %err-disp ...) handles ppc::error-fpu-exception-single
;; 03/12/97 bill  handle-stack-overflow properly updates sg.maxsize
;; 03/04/97 bill  handle-stack-overflow supports *minimum-stack-overflow-size*
;; 12/30/97 bill  handle-udf-call no longer calls bug.
;; 12/30/96 bill  handle-udf-call temporarily calls bug to display the symbol name
;; 10/29/96 bill  (defpascal %err-disp ...) handles ppc::error-too-many-values
;; -------------  4.0
;; 10/11/96 bill  handle-stack-overflow increases stack size while signalling error
;;                so that recursive overflows are less likely.
;; -------------  4.0f1
;; 06/26/96 bill  handle-stack-overflow comes out-of-line from (defpascal %err-disp ...).
;;                It determines if the stack size has gotten too large and signals
;;                a continuable error if it has.
;; 06/25/96 bill  handle-udf-call comes out of line from %err-disp. It makes restarting work.
;; -------------- MCL-PPC 3.9
;; 04/10/96 gb    handle ppc::error-memory-full error.
;; 04/08/96 gb    handle ppc::error-excised-function-call uuo.
;; 03/25/96 bill  (defpascal %err-disp ...) handles ppc::error-throw-tag-missing instead
;;                of the non-existent (to the kernel) ppc::error-divide-by-zero.
;; 03/01/96 gb    %err-disp callback handles (some) FPU exceptions
;; 02/22/96 bill  %kernel-restart-internal passes the frame-ptr as the first arg to
;;                the restart functions.
;; 01/31/96 gb    no need for nilreg-cell-symbol since 3.0.
;; 01/22/96 gb   %err-disp callback recognizes stack-overflow & alloc-failed.
;; 01/17/96 bill  Get frame-ptr from with-xp-stack-frames and call new %err-disp-internal
;;                or %kernel-restart-internal instead of %err-disp & %kernel-restart.
;; 12/08/95 bill  xp-gpr-lisp & friends move from here to ppc-trap-support
;; 11/30/95 bill  Remove the #_DebugStr from error.
;;                (defpascal %err-disp ...) now handles the UUO error codes correctly.
;; 11/29/95 bill  with-error-reentry-detection moves from here to l1-events
;; 11/21/95 bill (defpascal %err-disp ...) to handle callback from handle-error
;;               in lisp-exceptions.c. Wrap with-error-reentry-detection around
;;               the body of %err-disp to prevent infinite %err-disp loops.




(defun %stack-group-useable-size (&optional (sg *current-stack-group*))
  (multiple-value-bind (cf cu vf vu tf tu) (%stack-group-stack-space sg)
    (- (+ cf cu vf vu tf tu) (* 2 4096) *cs-hard-overflow-size* *cs-soft-overflow-size*)))

(defvar *allow-stack-overflows* t)

; This probably needs some tuning
(defvar *minimum-stack-overflow-size* (* 128 1024))



(queue-fixup
 (setq *allow-stack-overflows* nil))

(defun %kernel-restart (error-type &rest args)
  (%kernel-restart-internal error-type args (%get-frame-ptr)))

(defun %kernel-restart-internal (error-type args frame-ptr)
  ;(declare (dynamic-extent args))
  (dolist (f *kernel-restarts* (%err-disp-internal error-type args frame-ptr))
    (when (eq (car f) error-type)
      (return (apply (cdr f) frame-ptr args)))))

; this is the def of %err-disp.
; Yup.  That was my first guess.
(defun %err-disp (err-num &rest errargs)
  (%err-disp-internal err-num errargs (%get-frame-ptr)))

(defun %err-disp-internal (err-num errargs frame-ptr)
  (declare (fixnum err-num))
  ; The compiler (finally !) won't tail-apply error.  But we kind of
  ; expect it to ...
  (let* ((err-typ (max (ash err-num -16) 0))
         (err-num (%word-to-int err-num))
         (format-string (%rsc-string err-num))
         (condition-name (or (uvref *simple-error-types* err-typ)
                             (%cdr (assq err-num *kernel-simple-error-classes*)))))
    ;(dbg format-string)
    (if condition-name      
      (funcall '%error
               (case condition-name
                 (type-error (make-condition condition-name
                                             :format-string format-string
                                             :datum (car errargs)
                                             :expected-type (%type-error-type (cadr errargs))))
                 (file-error (make-condition condition-name
                                             :pathname (car errargs)
                                             :error-type format-string
                                             :format-arguments (cdr errargs)))
                 (undefined-function (make-condition condition-name
                                                     :name (car errargs)))
                 (t (make-condition condition-name 
                                    :format-string format-string
                                    :format-arguments errargs)))
               nil
               frame-ptr)
      (funcall '%error format-string errargs frame-ptr))))

(defun error (condition &rest args)
  #|
  #+ppc-target
  (with-pstrs ((pstr (if (stringp condition) condition "Error")))
    (#_DebugStr pstr))
  |#
  (%error condition args (%get-frame-ptr)))

(defun cerror (cont-string condition &rest args)
  (let* ((fp (%get-frame-ptr))
         (eval-queue *eval-queue*))
    (restart-case (%error condition (if (condition-p condition) nil args) fp)
      (continue ()
                :report (lambda (stream) 
                            (apply #'format stream cont-string args))
                (setq *eval-queue* eval-queue)
                nil))))

(defun %error (condition args error-pointer)
  (setq condition (condition-arg condition args 'simple-error))
  ;(with-pstrs ((str (format nil "error: ~a" condition))) (#_DebugStr str))  
  (signal condition)
  ;(with-pstrs ((str (format nil "error: ~a" condition))) (#_DebugStr str))
  (application-error *application* condition error-pointer)
  (application-error
   *application*
   (condition-arg "~s returned. It shouldn't.~%If it returns again, I'll throw to toplevel."
                  '(application-error) 'simple-error)
   error-pointer)
  (toplevel))
