;;;-*-Mode: LISP; Package: CCL -*-

;;	Change History (most recent first):
;;  13 9/13/96 akh  dont remember
;;  12 9/4/96  akh  maybe no change
;;  11 7/30/96 akh  fix find-str-pattern - dont run off end of string
;;  10 7/26/96 akh  remove support for old logical directories to fix some problems re back-translate-pathname
;;                  fix %path-str*=
;;  9 12/22/95 gb   bin;**;*.pfsl" maps to "ccl:binppc;**;*.pfsl
;;  7 11/15/95 akh  add to logical-pathname-translations for "CCL" so level-1;*.pfsl goes to l1-pfsls;
;;  4 10/11/95 akh  aargh
;;  5 2/21/95  slh  removed naughty word (file is public now)
;;  4 2/17/95  slh  added interface-tools, sourceserver modules
;;  (do not edit before this line!!)


;; L1-pathnames.lisp
; Copyright 1991-1994 Apple Computer, Inc.
; Copyright 1995 Digitool, Inc. The 'tool rules!

;; Modification History
;
; 11/16/95 bill "ccl:l1f;**;*.pfsl" -> "ccl:l1pf;**;*.pfsl"
; 2/16/95 slh   added interface tools, sourceserver to *module-search-path*
;---------------- 3.0d17
; %pathname-match-directory - do better with (null wild) 
;-------------------
;06/12/93 alice %path-str*= deals with fat strings
;09/30/92 alice translate component - source :unspecific, to "*" => result :unspecific
;05/19/92 alice translate-directory-2 - don't be so cavalier with to = **
;---------------- 2.0
;01/10/92 alice logical-pathname deal with :unspecific
;12/16/91 gb    string -> fixnum in signal-file-error call.
;---------------- 2.0b4
;10/06/91 alice add "ccl;" to translations - for top level
;---------------- 2.0b3
;07/21/91 gb    type-error -> signal-type-error.
;07/19/91 alice simpler better translate-directory2, fix recent bug in translate-component
;07/12/91 alice logical-pathname - make it work if host doesn't exist yet
;07/09/91 alice translate-pathname gets :reversible keyword (sp?)
;07/08/91 alice %split-ccdirectory had a bogus call to SET, translate-component deal with :wild
;06/10/91 alice  deal with :wild in directory lists (gaak)
;------------- 2.0b2
;05/20/91 gb    FILE-ERROR -> SIGNAL-FILE-ERROR.
;02/28/91 alice full-pathname - make no-error work if undefined logical host
;02/15/91 alice translations for e.g. lib;foo.fasl => bin;foo.fasl (add in boot too if we like it)
;02/13/91 alice added add-logical-pathname-translation 
;02/12/91 alice move tranlate-logical-pathname to end - add translations for ccl:
;02/07/91 alice - bunch of stuff moved from lib;pathnames

;-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_
;ANSI CL logical pathnames

(in-package :ccl)


;These are used by make-pathname
(defun %verify-logical-component (name type)
  (when (and name (neq name :unspecific))
    (setq name (ensure-simple-string name))
    (when (or (eql 0 (length name))
              (%str-member *pathname-escape-character* name)
              (%path-mem ":;" name))
      (error "Illegal logical pathname ~A component ~S" type name)))
  name)


(defun verify-logical-host-name (host)
  (or (and host (%verify-logical-component host "host")
       (%str-assoc host %logical-host-translations%)
       host)
      (host-error host)))

(defun %logical-version-component (version)
  (if (or (eq 0 version)
          (stringp version)
          (memq version '(nil :wild :newest :unspecific)))
    version
    (require-type version '(or (integer 0 0) string (member nil :wild :newest :unspecific)))))

(defun logical-pathname-translations (host)
  (setq host (verify-logical-host-name host))
  (let ((translations (%str-assoc host %logical-host-translations%)))
    (unless translations (host-error host))
    (%cdr translations)))

(defun logical-host-p (host)
  (%str-assoc host %logical-host-translations%))

(defun host-error (host) ; supposed to be a type-error
  (signal-type-error host  '(satisfies logical-host-p) "~S is not a defined logical host"))

(defun set-logical-pathname-translations (host list)
  (setq host (%verify-logical-component  host "host"))
  (let ((old (%str-assoc host %logical-host-translations%))
        new)
    (let ((%logical-host-translations% %logical-host-translations%))
      (cond ((null old) ; waste a cons to save some code - so we wont error cause of as yet undefined host
             (push (list host) %logical-host-translations%))
            ; not too cool if error in parse-namestring
            (t (rplacd old nil)))
      (setq new (mapcar #'(lambda (trans)
                            (destructuring-bind (from to &rest ignored) trans
                              (declare (ignore ignored))
                              (list (parse-namestring from host) (pathname to))))
                        list)))
    (dolist (e new)  ; ???
        (when (eq  (pathname-host (car e)) :unspecific)
          (error "~s is a physical pathname" (car e))))
    (if old
      (progn (%rplaca old host) (%rplacd old new))
      (push (cons host new) %logical-host-translations%)))
  list)

(defsetf logical-pathname-translations set-logical-pathname-translations)

; doest check if already there - adds at front 
(defun add-logical-pathname-translation (host translation)
  (let ((trans (%str-assoc host  %logical-host-translations%)))
    (if (not trans)
      (set-logical-pathname-translations host (list translation))
      (let ((new (debind (from to &rest ignored) translation
                        (declare (ignore ignored))
                        (list (parse-namestring from host) (pathname to)))))
        (rplacd trans (cons new (cdr trans)))
        (cdr trans)))))

(defun maybe-diddle-case (thing diddle-p)
  (if (and diddle-p (not (or (symbolp thing) (integerp thing))))
      (labels ((check-for (pred in)
		 (typecase in
		   (pattern
		    (dolist (piece (pattern-pieces in))
		      (when (typecase piece
			      (simple-string
			       (check-for pred piece))
			      (cons
			       (case (car in)
				 (:character-set
				  (check-for pred (cdr in))))))
			(return t))))
		   (list
		    (dolist (x in)
		      (when (check-for pred x)
			(return t))))
		   (simple-base-string
		    (dotimes (i (length in))
		      (when (funcall pred (schar in i))
			(return t))))
		   (t nil)))
	       (diddle-with (fun thing)
		 (typecase thing
		   (pattern
		    (make-pattern
		     (mapcar #'(lambda (piece)
				 (typecase piece
				   (simple-base-string
				    (funcall fun piece))
				   (cons
				    (case (car piece)
				      (:character-set
				       (cons :character-set
					     (funcall fun (cdr piece))))
				      (t
				       piece)))
				   (t
				    piece)))
			     (pattern-pieces thing))))
		   (list
		    (mapcar fun thing))
		   (simple-base-string 
		    (funcall fun thing))
		   (t
		    thing))))
	(let ((any-uppers (check-for #'upper-case-p thing))
	      (any-lowers (check-for #'lower-case-p thing)))
	  (cond ((and any-uppers any-lowers)
		 ;; Mixed case, stays the same.
		 thing)
		(any-uppers
		 ;; All uppercase, becomes all lower case.
		 (diddle-with #'(lambda (x) (if (stringp x)
						(string-downcase x)
						x)) thing))
		(any-lowers
		 ;; All lowercase, becomes all upper case.
		 (diddle-with #'(lambda (x) (if (stringp x)
						(string-upcase x)
						x)) thing))
		(t
		 ;; No letters?  I guess just leave it.
		 thing))))
      thing))

(defun substitute-into (pattern subs diddle-case)
  (declare (type pattern pattern)
	   (type list subs)
	   (values (or simple-base-string pattern)))
  (let ((in-wildcard nil)
	(pieces nil)
	(strings nil))
    (dolist (piece (pattern-pieces pattern))
      (cond ((simple-string-p piece)
	     (push piece strings)
	     (setf in-wildcard nil))
	    (in-wildcard)
	    (t
	     (setf in-wildcard t)
	     (unless subs
	       (error "Not enough wildcards in FROM pattern to match ~
		       TO pattern:~%  ~S"
		      pattern))
	     (let ((sub (pop subs)))
	       (typecase sub
		 (pattern
		  (when strings
		    (push (apply #'concatenate 'simple-string
				 (nreverse strings))
			  pieces))
		  (dolist (piece (pattern-pieces sub))
		    (push piece pieces)))
		 (simple-string
		  (push sub strings))
		 (t
		  (error "Can't substitute this into the middle of a word:~
			  ~%  ~S"
			 sub)))))))

    (when strings
      (push (apply #'concatenate 'simple-string (nreverse strings))
	    pieces))
    (values
     (maybe-diddle-case
      (if (and pieces (simple-string-p (car pieces)) (null (cdr pieces)))
	  (car pieces)
	  (make-pattern (nreverse pieces)))
      diddle-case)
     subs)))

(defun didnt-match-error (source from)
  (error "Pathname components from Source and From args to TRANSLATE-PATHNAME~@
	  did not match:~%  ~S ~S"
	 source from))

(defun compute-directory-substitutions (orig-source orig-from)
  (let ((source orig-source)
	(from orig-from))
    (let* ((subs ()))
      (loop
	(unless source
	  (unless (every #'(lambda (x) (eq x :wild-inferiors)) from)
	    (didnt-match-error orig-source orig-from))
	  (push () subs)
	  (return))
	(unless from (didnt-match-error orig-source orig-from))
	(let ((from-part (pop from))
	      (source-part (pop source)))
	  (typecase from-part
	    (pattern
	     (typecase source-part
	       (pattern
		(if (pattern= from-part source-part)
		    (push source-part subs)
		    (didnt-match-error orig-source orig-from)))
	       (simple-string
		(multiple-value-bind
		    (won new-subs)
		    (pattern-matches from-part source-part)
		  (if won
		      (dolist (sub new-subs)
			(push sub subs))
		      (didnt-match-error orig-source orig-from))))
	       (t
		(didnt-match-error orig-source orig-from))))
	    ((member :wild)
	     (push source-part subs))
	    ((member :wild-inferiors)
	     (let ((remaining-source (cons source-part source)))
	       (let* ((res ()))
		 (loop
		   (when (directory-components-match remaining-source from)
		     (return))
		   (unless remaining-source
		     (didnt-match-error orig-source orig-from))
		   (push (pop remaining-source) res))
		 (push (nreverse res) subs)
		 (setq source remaining-source))))
	    (simple-string
	     (unless (and (simple-string-p source-part)
			  (string= from-part source-part))
	       (didnt-match-error orig-source orig-from)))
	    (t
	     (didnt-match-error orig-source orig-from)))))
      (nreverse subs))))

(defun translate-component (source from to diddle-case) 
  (typecase to
    (pattern
     (typecase from
       (pattern
	(typecase source
	  (pattern
	   (if (pattern= from source)
	       source
	       (didnt-match-error source from)))
	  (simple-string
	   (multiple-value-bind
	       (won subs)
	       (pattern-matches from source)
	     (if won
		 (values (substitute-into to subs diddle-case))
		 (didnt-match-error source from))))
	  (t
	   (maybe-diddle-case source diddle-case))))
       ((member :wild)
	(values (substitute-into to (list source) diddle-case)))
       (t
	(if (components-match source from)
	    (maybe-diddle-case source diddle-case)
	    (didnt-match-error source from)))))
    ((member nil :wild)
     (maybe-diddle-case source diddle-case))
    (t
     (if (components-match source from)
	 to
	 (didnt-match-error source from)))))

(defun translate-directories (source from to diddle-case)
  (if (not (and #|source|# to from))
      (or to
	  (mapcar #'(lambda (x) (maybe-diddle-case x diddle-case)) source))
      (let* ((res ()))
	(push (first (or source from)) res)
	(let ((subs-left (compute-directory-substitutions (rest source)
							  (rest from))))
	  (dolist (to-part (rest to))
	    (typecase to-part
	      ((member :wild)
	       (assert subs-left)
	       (let ((match (pop subs-left)))
		 (when (listp match)
		   (error ":WILD-INFERIORS not paired in from and to ~
			   patterns:~%  ~S ~S" from to))
		 (maybe-diddle-case match diddle-case)))
	      ((member :wild-inferiors)
	       (let ((match (pop subs-left)))
		 (unless (listp match)
		   (error ":WILD-INFERIORS not paired in from and to ~
			   patterns:~%  ~S ~S" from to))
		 (dolist (x match)
		   (push (maybe-diddle-case x diddle-case) res))))
	      (pattern
	       (multiple-value-bind
		   (new new-subs-left)
		   (substitute-into to-part subs-left diddle-case)
		 (setf subs-left new-subs-left)
		 new))
	      (t (push to-part res)))))
	(nreverse res))))

(defun translate-pathname (source from-wildname to-wildname &key)
  "Use the source pathname to translate the from-wildname's wild and
   unspecified elements into a completed to-pathname based on the to-wildname."
  (declare (type path-designator source from-wildname to-wildname))
  (let* ((source (pathname source)))
    (let* ((from (pathname from-wildname)))
      (let* ((to (pathname to-wildname)))
	  (let* ((source-host (pathname-host source))
		 (to-host (pathname-host to))
		 (diddle-case nil))
	    (macrolet ((frob (field &optional (op 'translate-component))
			 `(let ((result (,op (,field source)
					     (,field from)
					     (,field to)
					     diddle-case)))
			    (if (eq result :error)
				(error "~S doesn't match ~S" source from)
				result))))
	      (cons-pathname
	       (frob pathname-directory translate-directories)
	       (frob %pathname-name)
	       (frob %pathname-type)
	       (or to-host source-host)
	       (frob pathname-version))))))))












(defun logical-pathname (path)
  (let* (host
         (result
         (typecase path
           (logical-pathname path)
           (stream (%path-from-stream path))
           (string 
            (setq host (pathname-host-sstr path 0 (length path) t))))))
    (if (and host (neq host :unspecific) (null (%str-assoc host %logical-host-translations%)))
      (setf (logical-pathname-translations host) nil))
    (typecase result
      (logical-pathname result)
      (string (pathname path))
      (t (report-bad-arg path 'logical-pathname))))) ; huh? - a tautology


(defun pattern= (pattern1 pattern2)
  (declare (type pattern pattern1 pattern2))
  (let ((pieces1 (pattern-pieces pattern1))
	(pieces2 (pattern-pieces pattern2)))
    (and (= (length pieces1) (length pieces2))
	 (every #'(lambda (piece1 piece2)
		    (typecase piece1
		      (simple-string
		       (and (simple-string-p piece2)
			    (string= piece1 piece2)))
		      (cons
		       (and (consp piece2)
			    (eq (car piece1) (car piece2))
			    (string= (cdr piece1) (cdr piece2))))
		      (t
		       (eq piece1 piece2))))
		pieces1
		pieces2))))

(defun pattern-matches (pattern string)
  (declare (type pattern pattern)
	   (type simple-string string))
  (let ((len (length string)))
    (labels ((maybe-prepend (subs cur-sub chars)
	       (if cur-sub
		   (let* ((len (length chars))
			  (new (make-string len :element-type 'base-char))
			  (index len))
		     (dolist (char chars)
		       (setf (schar new (decf index)) char))
		     (cons new subs))
		   subs))
	     (matches (pieces start subs cur-sub chars)
	       (if (null pieces)
		   (if (= start len)
		       (values t (maybe-prepend subs cur-sub chars))
		       (values nil nil))
		   (let ((piece (car pieces)))
		     (etypecase piece
		       (simple-string
			(let ((end (+ start (length piece))))
			  (and (<= end len)
			       (string= piece string
					:start2 start :end2 end)
			       (matches (cdr pieces) end
					(maybe-prepend subs cur-sub chars)
					nil nil))))
		       (list
			(ecase (car piece)
			  (:character-set
			   (and (< start len)
				(let ((char (schar string start)))
				  (if (find char (cdr piece) :test #'char=)
				      (matches (cdr pieces) (1+ start) subs t
					       (cons char chars))))))))
		       ((member :single-char-wild)
			(and (< start len)
			     (matches (cdr pieces) (1+ start) subs t
				      (cons (schar string start) chars))))
		       ((member :multi-char-wild)
			(multiple-value-bind
			    (won new-subs)
			    (matches (cdr pieces) start subs t chars)
			  (if won
			      (values t new-subs)
			      (and (< start len)
				   (matches pieces (1+ start) subs t
					    (cons (schar string start)
						  chars)))))))))))
      (multiple-value-bind
	  (won subs)
	  (matches (pattern-pieces pattern) 0 nil nil nil)
	(values won (reverse subs))))))

(defun components-match (thing wild)
  (declare (type (or pattern symbol simple-string integer) thing wild))
  (or (eq thing wild)
      (eq wild :wild)
      (typecase thing
	(simple-base-string
	 ;; String is matched by itself, a matching pattern or :WILD.
	 (typecase wild
	   (pattern 
	    (values (pattern-matches wild thing)))
	   (simple-base-string
	    (string= thing wild))))
	(pattern
	 ;; A pattern is only matched by an identical pattern.
	 (and (pattern-p wild) (pattern= thing wild)))
	(integer
	 ;; an integer (version number) is matched by :WILD or the same
	 ;; integer.  This branch will actually always be NIL as long is the
	 ;; version is a fixnum.
	 (eql thing wild)))))

(defun directory-components-match (thing wild)
  (or (eq thing wild)
      (eq wild :wild)
      (and (consp wild)
	   (or (and (= (length wild) 2)
		    (eq (car wild) :absolute)
		    (null thing)
		    (eq (cadr wild) :wild-inferiors))
	       (let ((wild1 (first wild)))
		 (if (eq wild1 :wild-inferiors)
		     (let ((wild-subdirs (rest wild)))
		       (or (null wild-subdirs)
			   (loop
			       (when (directory-components-match thing wild-subdirs)
				 (return t))
			       (pop thing)
			     (unless thing (return nil)))))
		     (and (consp thing)
			  (components-match (first thing) wild1)
			  (directory-components-match (rest thing)
						      (rest wild)))))))))


(defun pathname-match-p (in-pathname in-wildname)
  "Pathname matches the wildname template?"
  (declare (type path-designator in-pathname))
  (let* ((pathname (pathname in-pathname)))
    (let* ((wildname (pathname in-wildname)))
      (macrolet ((frob (field &optional (op 'components-match ))
		   `(or (null (,field wildname))
			(,op (,field pathname) (,field wildname)))))
	(and (or (null (pathname-host wildname))
		 (equal (pathname-host wildname) (pathname-host pathname)))
	     (frob pathname-directory directory-components-match)
	     (frob pathname-name)
	     (frob pathname-type)
	     (frob pathname-version))))))









(defun user-homedir-pathname (&optional host)
  (declare (ignore host))  
  (let* ((dir (get-user-home-dir (getuid))))
    (if dir
	(make-pathname :defaults nil :directory dir))))

(progn
(setf (logical-pathname-translations "home")
          `(("*.*" ,(merge-pathnames "*.*" (user-homedir-pathname)))))

(setf (logical-pathname-translations "ccl")
          `(("lib;**;*.fasl" "ccl:bin;*.fasl")
	    ("l1;**;*.fasl" "ccl:l1f;*.fasl")
            ("l1;**;*.pfsl" "ccl:l1pf;*.pfsl")
            ("l1;**;*.sfsl" "ccl:l1sf;*.sfsl")
            ("l1;**;*.*" "ccl:level-1;**;*.*")
            ("l1f;**;*.pfsl" "ccl:l1pf;**;*.pfsl")
            ("l1f;**;*.sfsl" "ccl:l1sf;**;*.sfsl")
            ("bin;**;*.pfsl" "ccl:binppc;**;*.pfsl")
            ("bin;**;*.sfsl" "ccl:binsparc;**;*.sfsl")
            ("l1pf;**;*.*" "ccl:l1-pfsls;**;*.*")
            ("l1sf;**;*.*" "ccl:l1-sfsls;**;*.*")
            ("l1f;**;*.*" "ccl:l1-fasls;**;*.*")
            ("ccl;*.*" ,(merge-pathnames "*.*" (mac-default-directory)))
            ("**;*.*" ,(merge-pathnames "**/*.*" (mac-default-directory)))))

)

(defloadvar *user-homedir-pathname* (user-homedir-pathname))

(defun translate-logical-pathname (pathname &key)
  (setq pathname (pathname pathname))
  (let ((host (pathname-host pathname)))
    (if (or (null host) (eq host :unspecific))
      pathname
      (let ((rule (assoc pathname (logical-pathname-translations host)
                         :test #'pathname-match-p)))  ; how can they match if hosts neq??
        (if rule
          (translate-logical-pathname
           (translate-pathname pathname (car rule) (cadr rule)))
          (signal-file-error $xnotranslation pathname))))))

;This function should be changed to standardize the name more than it does.
;It should eliminate non-leading instances of "::" etc at least.  We might also
;want it to always return an absolute pathname (i.e. fill in the default mac
;directory), so as to make it a sort of harmless truename (which is how I think
;it's mainly used).  Unfortunately that would make it go to the file system,
;but it might be worth it.
;This function used to also remove quoting so as to make the name suitable for
;passing to rom.  It doesn't anymore. Use mac-namestring for that.
; does anybody use this??
; DO - merge in default if relative, and do the :: stuff
; perhaps call it expand-pathname or expanded-pathname


(defun full-pathname (path &key (no-error t))
  (cond (no-error
         ; note that ignore-errors wont work until var %handlers% is defined (in l1-init)
         (setq path (ignore-errors
                     (translate-logical-pathname (merge-pathnames path))))
         (when (null path) (return-from full-pathname nil)))
        (t (setq path (translate-logical-pathname (merge-pathnames path)))))
  (let ((dir (%pathname-directory path)))
    (when (and no-error (not dir) (%pathname-directory path)) ; WHAT is  that noop - since 3.0??
      (return-from full-pathname nil))
    (setq dir (absolute-directory-list dir))
    (unless (eq dir (%pathname-directory path))
      (setq path (cons-pathname dir (%pathname-name path) (%pathname-type path)
                                (pathname-host path)(pathname-version path))))
  path))







(defparameter *module-search-path* (list
                                    (cons-pathname '(:absolute "bin") nil nil "ccl")
                                    *user-homedir-pathname*
                                    (cons-pathname  nil nil nil "ccl")
                                    (cons-pathname '(:absolute "lib") nil nil "ccl")
				    (cons-pathname '(:absolute "library") nil nil "ccl"))
  "Holds a list of pathnames to search for the file that has same name
   as a module somebody is looking for.")

#|
	Change History (most recent last):
	2	12/29/94	akh	merge with d13
	3	1/2/95	akh	actually no change
|# ;(do not edit past this line!!)
