;;;-*-Mode: LISP; Package: CCL -*-

;;	Change History (most recent first):
;;  11 1/22/97 akh  print-listener-prompt updates edit-menu iff disabled.
;;  10 10/3/96 akh  find-restart-in-process to l1-readloop
;;  3 10/17/95 akh  merge patch to print-listener-prompt
;;  2 10/13/95 bill ccl3.0x25
;;  12 5/23/95 akh  add a comment
;;  11 5/22/95 akh  close restart windows when closing backtrace window
;;                  Add application-eval-enqueue method
;;                  Make eval-enqueue work from modal dialog
;;  5 4/10/95  akh  fboundp => boundp in eval-enqueue
;;  3 4/6/95   akh  make eval-enqueue work in init file
;;  (do not edit before this line!!)


; l1-readloop-lds.lisp
; Copyright 1995 Digitool, Inc. The 'tool rules!

; Modification History
;
; 03/04/97 bill  break-loop ensures that the backtrace-info vector
;                does not have any null stack frame entries.
; 01/15/97 bill  %break-message unconditionally binds *signal-printing-errors* to NIL.
; 11/19/96 bill  slh's change to cbreak-loop to stop it from
;                popping up an empty listener when *autoclose-inactive-listeners*
;                it true.
; 12/09/96 akh  print-listener-prompt updates edit-menu iff disabled.
; -------------  4.0
;;<end of added text>
; 07/20/96 bill  dummy definitions of databases-locked-p & funcall-with-databases-unlocked.
;                Real definition of with-databases-unlocked.
;                Add database lock detection to break-loop.
; 06/04/96 bill  in read-loop - Initial binding of *eval-queue* comes from global value.
; -------------  MCL-PPC 3.9
; 04/01/96 bill  interactive-break binds *interrupt-level* to 0 to avoid
;                confusing break loop message.
; 01/17/96 gb    bind *interrupt-level* to 0 vice 2.
; 01/16/96 bill  on the PPC, break-loop-handle-error doesn't do parent-frame
;                before passing the error-pointer to break-loop.
; 12/05/95 bill  break-loop sets *error-reentry-count* to 0.
; 10/11/95 bill  %break-message doesn't print the standin event processor
;                if no standin event processor was created.
;                read-loop sets *processing-events* to NIL if it is non-NIL
;                and the current process is the event processor.
;                toplevel-loop continues throwing to the real toplevel
;                if the current process is the initial process.
;  6/08/95 slh   help-specs
;  5/05/95 slh   break-loop: "is no longer" -> "may no longer be" (because a
;                break-on-signal loop can exit on cmd-/ to a break-on-error loop)
;                "Dead" -> "inactive"; added *autoclose-inactive-listeners*
;                %break-message: space after prefix char
;  4/26/95 slh   toplevel-read: bail out of process if listener is unuseable
;                break-loop: check for listener process in unwind-cleanup
;  4/24/95 slh   break: obey *break-loop-when-uninterruptable*
;  4/18/95 slh   require inspector done inline
;  4/05/95 slh   select-backtrace: require 'new-backtrace

(in-package "CCL")



  




(defun continuable-process-p ()
  (without-interrupts
   (dolist (p *active-processes*)
     (let ((sg (process-stack-group p)))
       (when (symbol-value-in-stack-group '*continuablep* sg)
         (return p))))))

(defun continuable-processes ()
  (without-interrupts
   (let ((result))
     (dolist (process *active-processes*)
         (when  (symbol-value-in-process '*continuablep* process)
           (push process result)))
     result)))

; this isnt quite right but requiring *in-read-loop* isnt right either.
(defun restartable-process-p ()
  (without-interrupts
   (dolist (p *active-processes*)
     (when (neq 0 (symbol-value-in-process '*break-level* p))
       (return p)))))

(defun restartable-processes ()
  (without-interrupts
   (let ((result))
     (dolist (process *active-processes*)
       (when (neq 0 (symbol-value-in-process '*break-level* process))
         (push process result)))
     result)))

(defun process-to-continue ()
  (let ((ps (continuable-processes)))
    (case (length ps)
      (0 nil)
      (1 (car ps))
      (t (let ((p (catch-cancel (select-item-from-list ps :window-title "Continue Process"))))
           (if (neq p :cancel) (car p)))))))

(defun process-to-restart (&optional (string "Restart Process"))
  (let ((ps (restartable-processes)))
    (case (length ps)
      (0 nil)
      (1 (car ps))
      (t (let ((p (catch-cancel (select-item-from-list ps :window-title string))))
           (if (neq p :cancel) (car p)))))))


(defun interactive-break ()
  (let* ((w (front-window))
         (p (if w (window-process w))))
    (when (or (not p)(neq p *current-process*)) (setq p (process-to-abort "Break in Process")))
    (if p (process-interrupt p 
                             #'(lambda ()
                                 (let ((*interrupt-level* 0))
                                   (break)))))))

(defun toplevel-loop ()
  (loop
    (if (eq (catch :toplevel 
              (read-loop 0)) $xstkover)
      (format t "~&;[Stacks reset due to overflow.]")
      (when (eq *current-process* *initial-process*)
        (toplevel)))))

;This is the part common to toplevel loop and inner break loops.
(defun read-loop (level)
  (declare (resident))
  "Never returns"
  (read-loop-internal level))

(defvar *toplevel-commands* ())

(defun %define-toplevel-command (key name fn doc args)
  (let* ((pair (assoc key *toplevel-commands* :test #'eq)))
    (if pair
      (rplacd pair (list* fn doc args))
      (push (cons key (list* fn doc args)) *toplevel-commands*)))
  name)
  
(define-toplevel-command pop () "exit current break loop" (abort-break))
(define-toplevel-command go () "continue" (continue))
(define-toplevel-command q () "return to toplevel" (toplevel))
(define-toplevel-command r () "list restarts"
  (let* ((r (apply #'vector (compute-restarts *break-condition*))))
    (dotimes (i (length r) (terpri))
      (format t "~&~d. ~a" i (svref r i)))))

(define-toplevel-command ? () "help"
  (dolist (c *toplevel-commands*)
    (let* ((command (car c))
           (doc (caddr c))
           (args (cdddr c)))
      (if args
         (format t "~& (~S~{ ~A~}) ~8T~A" command args doc)
         (format t "~& ~S  ~8T~A" command doc)))))

(define-toplevel-command b (&optional show-frame-contents) "backtrace"
  (when *break-frame*
    (print-call-history :detailed-p show-frame-contents
                        :start-frame *break-frame*)))

(define-toplevel-command c (n) "Choose restart <n>"
   (select-restart n))

(define-toplevel-command f (n) "Show backtrace frame <n>"
   (print-call-history :start-frame *break-frame*
                       :detailed-p n))

(defun check-toplevel-command (form)
  (let* ((cmd (if (consp form) (car form) form))
         (args (if (consp form) (cdr form))))
    (if (keywordp cmd)
      (let* ((pair (assoc cmd *toplevel-commands*)))
	(if pair 
	    (progn (apply (cadr pair) args)
		   t))))))

(defun read-loop-internal (level)
  (let* ((*break-level* level)
         (*last-break-level* level)
         *loading-file-source-file*
         *in-read-loop*
         (*listener-p* t)
         *** ** * +++ ++ + /// // / -
         form)
    (loop
      (restart-case
        (catch :abort ;last resort...
          (loop
            (catch-cancel
              (loop                
                (setq *loading-file-source-file* nil
                      *in-read-loop* nil
                      *break-level* level)
                (setq form (toplevel-read))
                (if (eq form *eof-value*)
		  (progn
		    (stream-clear-input *terminal-io*)
		    (abort-break))
                  (with-non-background-process
		    (or (check-toplevel-command form)
			(toplevel-print
			 (toplevel-eval form)))))))
            (format *terminal-io* "~&Cancelled")))
        (abort () :report (lambda (stream)
                            (if (eq level 0)
                              (format stream "Return to toplevel.")
                              (format stream "Return to break level ~D." level)))
               #| ; Handled by interactive-abort
                ; go up one more if abort occurred while awaiting/reading input               
                (when (and *in-read-loop* (neq level 0))
                  (abort))
                |#
               )
        (abort-break () 
                     (unless (eq level 0)
                       (abort))))
      (clear-input *terminal-io*)
      (format *terminal-io* "~%"))))

;Read a form from *terminal-io*.
(defun toplevel-read ()
  (declare (resident))
  (let* ((listener *terminal-io*))
    (declare (dynamic-extent wait-function))
    (force-output *standard-output*)
    (print-listener-prompt t)
    (loop
      (when t ;(process-wait-with-timeout "Input" 10 wait-function)
        (let* ((*in-read-loop* nil)  ;So can abort out of buggy reader macros...
               (form))
          (catch '%re-read            
            (if (eq (setq form (read listener nil *eof-value*)) *eof-value*)
              (progn
                (return form))
	      (progn
		(let ((ch)) ;Trim whitespace
		  (while (and (listen listener) (whitespacep (setq ch (tyi listener))))
		    (setq ch nil))
		  (when ch (untyi ch listener)))
		(when *listener-indent* 
		  (tyo #\space listener)
		  (tyo #\space listener))
		(return (process-single-selection form))))))))))

(defvar *always-eval-user-defvars* nil)

(defun process-single-selection (form)
  (if (and *always-eval-user-defvars*
           (listp form) (eq (car form) 'defvar) (cddr form))
    `(defparameter ,@(cdr form))
    form))

(defun toplevel-eval (form &optional env &aux values)
   (declare (resident))
  (setq +++ ++ ++ + + - - form)
  (setq values (multiple-value-list (cheap-eval-in-environment form env)))
  values)

(defun toplevel-print (values)
  (declare (resident))
  (setq /// // // / / values)
  (setq *** ** ** * * (if (neq (%car values) (%unbound-marker-8)) (%car values)))
  (when values
    (fresh-line)
    (dolist (val values) (write val) (terpri))))

(defun print-listener-prompt (&optional force)
  (when (or force (neq *break-level* *last-break-level*))
    (let* ((*listener-indent* nil))
      (fresh-line *terminal-io*)            
      (if (%izerop *break-level*)
        (%write-string "?" *terminal-io*)
        (format *terminal-io* "~s >" *break-level*)))        
    (write-string " " *terminal-io*)        
    (setq *last-break-level* *break-level*))
      (force-output *terminal-io*))

; You may want to do this anyway even if your application
; does not otherwise wish to be a "lisp-development-system"
(defmethod application-error ((a lisp-development-system) condition error-pointer)
  (break-loop-handle-error condition error-pointer))

(defun break-loop-handle-error (condition error-pointer)
  (multiple-value-bind (bogus-globals newvals oldvals) (%check-error-globals)
    (dolist (x bogus-globals)
      (set x (funcall (pop newvals))))
    (when (and *debugger-hook* *break-on-errors*)
      (let ((hook *debugger-hook*)
            (*debugger-hook* nil))
        (funcall hook condition hook)))
    (%break-message (error-header "Error") condition error-pointer)
    (let* ((s *error-output*))
      (dolist (bogusness bogus-globals)
        (let ((oldval (pop oldvals)))
          (format s "~&;  NOTE: ~S was " bogusness)
          (if (eq oldval (%unbound-marker-8))
            (format s "unbound")
            (format s "~s" oldval))
          (format s ", was reset to ~s ." (symbol-value bogusness)))))
    (if *break-on-errors*
      (break-loop condition error-pointer)
      (abort))))

(defun break (&optional string &rest args &aux (fp (%get-frame-ptr)))
  (flet ((do-break-loop ()
           (let ((c (make-condition 'simple-condition
                                    :format-string (or string "")
                                    :format-arguments args)))
             (cbreak-loop (error-header "Break") "Return from BREAK." c fp))))
    (cond ((%i> *interrupt-level* -1)
           (do-break-loop))
          (*break-loop-when-uninterruptable*
           (format *error-output* "Break while *interrupt-level* less than zero; binding to 0 during break-loop.")
           (let ((*interrupt-level* 0))
             (do-break-loop)))
          (t (format *error-output* "Break while *interrupt-level* less than zero; ignored.")))))

(defun invoke-debugger (&optional string &rest args &aux (fp (%get-frame-ptr)))
  (let ((c (condition-arg (or string "") args 'simple-condition)))
    (when *debugger-hook*
      (let ((hook *debugger-hook*)
            (*debugger-hook* nil))
        (funcall hook c hook)))
    (%break-message "Debug" c fp)
    (break-loop c fp)))

(defun %break-message (msg condition error-pointer &optional (prefixchar #\>))
  (let ((*print-circle* *error-print-circle*)
        ;(*print-pretty* nil)
        (*print-array* nil)
        (*print-escape* t)
        (*print-gensym* t)
        (*print-length* nil)  ; ?
        (*print-level* nil)   ; ?
        (*print-lines* nil)
        (*print-miser-width* nil)
        (*print-readably* nil)
        (*print-right-margin* nil)
        (*signal-printing-errors* nil)
        (s (make-indenting-string-output-stream prefixchar nil)))
    (format s "~A ~A: " prefixchar msg)
    (setf (indenting-string-output-stream-indent s) (column s))
    ;(format s "~A" condition) ; evil if circle
    (report-condition condition s)
    (if (not (and (typep condition 'simple-program-error)
                  (simple-program-error-context condition)))
      (format *error-output* "~&~A~%~A While executing: ~S~%"
              (get-output-stream-string s) prefixchar (%real-err-fn-name error-pointer))
      (format *error-output* "~&~A~%"
              (get-output-stream-string s)))
  (force-output *error-output*)))   ; returns NIL

(defun cbreak-loop (msg cont-string condition error-pointer)
  (let* ((*print-readably* nil))
    (%break-message msg condition error-pointer)
    (restart-case (break-loop condition error-pointer *backtrace-on-break*)
      (continue () :report (lambda (stream) (write-string cont-string stream))))
    (fresh-line *error-output*)
    nil))

(defun warn (format-string &rest args)
  (let ((fp (%get-frame-ptr))
        (c (require-type (condition-arg format-string args 'simple-warning) 'warning)))
    (when *break-on-warnings*
      (cbreak-loop "Warning" "Signal the warning." c fp))
    (restart-case (signal c)
      (muffle-warning () :report "Skip the warning" (return-from warn nil)))
    (%break-message (if (typep c 'compiler-warning) "Compiler warning" "Warning") c fp #\;)
    ))

(declaim (notinline select-backtrace))

(defmacro new-backtrace-info (dialog youngest oldest stack-group)
  `(vector ,dialog ,youngest ,oldest ,stack-group nil))

(defun select-backtrace ()
  (declare (notinline select-backtrace))
  (require 'new-backtrace)
  (require :inspector)
  (select-backtrace))

(defvar *break-condition* nil "condition argument to innermost break-loop.")
(defvar *break-frame* nil "frame-pointer arg to break-loop")
(defvar *break-loop-when-uninterruptable* t)




(eval-when (:compile-toplevel :execute :load-toplevel)

(unless (fboundp 'databases-locked-p)

; Redefined by Wood
(defun databases-locked-p (&optional by-locker)
  (declare (ignore by-locker))
  nil)

)  ; end of unless

(unless (fboundp 'funcall-with-databases-unlocked)

; Redefined by Wood
(defun funcall-with-databases-unlocked (thunk)
  (funcall thunk))

))  ; end of unless and eval-when

(defmacro with-databases-unlocked (&body body)
  (let ((thunk (gensym)))
    `(let ((,thunk #'(lambda () ,@body)))
       (declare (dynamic-extent ,thunk))
       (funcall-with-databases-unlocked ,thunk))))

(defvar %last-continue% nil)
(defun break-loop (condition frame-pointer
                             &optional (backtracep *backtrace-on-break*))
  "Never returns"
  (when (and (%i< *interrupt-level* 0) (not *break-loop-when-uninterruptable*))
    (abort))
  (let* ((%handlers% nil)           ; firewall
         (databases-locked-p (databases-locked-p *current-process*))
         (*break-frame* frame-pointer)
         (*break-condition* condition)
         (*compiling-file* nil)
         (*backquote-stack* nil)
         (continue (find-restart 'continue))
         (*continuablep* (unless (eq %last-continue% continue) continue))
         (%last-continue% continue)
         (*standard-input* *debug-io*)
         (*standard-output* *debug-io*)
         (*interrupt-level* 0)
         (*signal-printing-errors* nil)
         (*read-suppress* nil)
         (*print-readably* nil))
    (with-databases-unlocked
      (progn
        (if *continuablep*
          (let* ((*print-circle* *error-print-circle*)
                 ;(*print-pretty* nil)
                 (*print-array* nil))
            (format t "~&> Type :GO to continue, :POP to abort.")
            (format t "~&> If continued: ~A~%" continue))
          (format t "~&> Type :POP to abort.~%"))
        (format t "~&Type :? for other options.")
        (terpri)
        (when databases-locked-p
          (format t "> WARNING: The WOOD database lock was locked.~@
                           > Use caution or you might damage your database.~%"))
        (force-output)
        (when backtracep
          (select-backtrace))
        (clear-input *debug-io*)
        (setq *error-reentry-count* 0)        ; succesfully reported error
        (read-loop (1+ *break-level*))))))



; unused?
(defun display-restarts (&optional (condition *break-condition*))
  (let ((i 0))
    (format t "~&[Pretend that these are buttons.]")
    (dolist (r (compute-restarts condition) i)
      (format t "~&~a : ~A" i r)
      (setq i (%i+ i 1)))
    (fresh-line nil)))

(defun select-restart (n &optional (condition *break-condition*))
  (let* ((restarts (compute-restarts condition)))
    (invoke-restart-interactively
     (nth (require-type n `(integer 0 (,(length restarts)))) restarts))))




; End of l1-readloop-lds.lisp
