;;;-*- Mode: Lisp; Package: CCL -*-

;;	Change History (most recent first):
;;  1 11/16/95 bill New file
;;  (do not edit before this line!!)


;;; ppc-callback-support.lisp
;;;
;;; Support for PPC callbacks

;;; Modification History
;;;
;;; 03/10/97 bill  A version of the backward-compatible define-ppc-pascal-function
;;;                that actually works.
;;; 03/09/97 gb    define-ppc-pascal-function -> define-ppc-pascal-function-2, backward compat.
;;; 02/23/97 bill  %pascal-functions% now passes the stack-ptr-fixnum to the
;;;                lisp function. This prevents VSP use for lisp functions
;;;                that have no args (cs-overflow-callback makes use of this fact).
;;; -------------  4.0
;;; 10/07/96 bill  #+ppc-target version of defpascal-callback-p
;;;  9/06/96 slh   define-ppc-pascal-function: record 'defpascal (from AlanR)
;;; 07/31/96 bill  new upp-transition-vector function
;;; -------------  MCL-PPC 3.9
;;; 04/03/96 bill  %cons-pfe takes a new, without-interrupts arg.
;;;                define-ppc-pascal-function takes a new, optional without-interrupts
;;;                arg, which it puts into the generated pfe.
;;;                %pascal-functions% is optimized (speed 3) (safety 0) so that
;;;                it won't do event check on entry. It conditionally executes
;;;                it's body without-interrupts according to the pfe.without-interrupts
;;;                value.
;;; 03/25/96 bill  define-ppc-pascal-function takes an optional doc-string arg
;;; 02/22/96 bill *call-universal-proc-slep*, *call-universal-proc-address*
;;; 01/23/96 bill  %fixnum-from-macptr (opposite of %setf-macptr-to-object).
;;; 01/16/96 bill  define-ppc-pascal-function correctly initializes new entries in
;;;                %pascal-functions to NIL. It also doesn't attempt to %svref
;;;                entries that are not vectors.
;;; 11/21/95 bill  New %get-object function. Add one instruction memory access time in %set-object.
;;; 11/10/95 bill  New file
;;;

;;; Universal Procedure Pointer -> Transition Vector = (address, TOC, ...)
;;; The transition vector for our callbacks looks like:
;;;
;;; .SPcallback      ; Subprim that trampolines back to the Lisp
;;; nil-ptr          ; pointer to the nil below
;;; nil              ; for initializing nilreg
;;; index            ; Index in %pascal-functions% of lisp code. Boxed.
;;;
;;; sp-callback uses the nil-ptr value it finds in the TOC register (nilreg)
;;; to load nilreg and get the %pascal-functions% index. Then it
;;; funcalls #'%pascal-functions% with two args, the %pascal-functions% index
;;; and a pointer to the stack frame containing the arguments (tagged as a fixnum).
;;; %pascal-functions% puts the return value in param0.

  ; end eval-when




; Return as a fixnum the address of the subprim at the given offset
#+ppc-target
(defppclapfunction %get-subprim ((subprim-offset arg_z))
  (ref-global imm0 subprims-base)
  (unbox-fixnum arg_z arg_z)
  (add arg_z imm0 arg_z)
  (blr))


#+ppc-target
(defppclapfunction %get-object ((macptr arg_y) (offset arg_z))
  (twnei nargs (* 2 4))
  (trap-unless-typecode= arg_y ppc::subtag-macptr)
  (macptr-ptr imm0 arg_y)
  (trap-unless-lisptag= arg_z ppc::tag-fixnum imm1)
  (unbox-fixnum arg_z arg_z)
  (lwzx arg_z arg_z imm0)
  (blr))

;; It would be awfully nice if (setf (%get-long macptr offset)
;;                                   (ash (the fixnum value) ppc::fixnumshift))
;; would do this inline.
#+ppc-target
(defppclapfunction %set-object ((macptr arg_x) (offset arg_y) (value arg_z))
  (twnei nargs (* 3 4))
  (trap-unless-typecode= arg_x ppc::subtag-macptr)
  (macptr-ptr imm0 arg_x)
  (trap-unless-lisptag= arg_y ppc::tag-fixnum imm1)
  (unbox-fixnum arg_y arg_y)
  (stwx arg_z arg_y imm0)
  (blr))

;; It would be nice if (%setf-macptr macptr (ash (the fixnum value) ppc::fixnumshift))
;; would do this inline.
#+ppc-target
(defppclapfunction %setf-macptr-to-object ((macptr arg_y) (object arg_z))
  (twnei nargs (* 2 4))
  (trap-unless-typecode= arg_y ppc::subtag-macptr)
  (stw arg_z ppc::macptr.address arg_y)
  (blr))

(defppclapfunction %fixnum-from-macptr ((macptr arg_z))
  (check-nargs 1)
  (trap-unless-typecode= arg_z ppc::subtag-macptr)
  (lwz imm0 ppc::macptr.address arg_z)
  (trap-unless-lisptag= imm0 ppc::tag-fixnum imm1)
  (mr arg_z imm0)
  (blr))


;; This is machine-dependent (it conses up a piece of "trampoline" code
;; which calls a subprim in the lisp kernel.)
(defun make-callback-trampoline (index)
  (macrolet ((ppc-lap-word (instruction-form)
               (uvref (uvref (compile nil `(lambda (&lap 0) (ppc-lap-function () ((?? 0)) ,instruction-form))) 0) 0)))
    (let* ((nil-address (%address-of nil))
           (p (malloc 24)))
      (setf (%get-long p 0) (logior (ash nil-address -16) 
                                    (ppc-lap-word (lis 12 ??)))   ; high half of nil address
            (%get-long p 4) (logior (ppc-lap-word (ori 12 12 ??))
                                    (logand #xffff nil-address))        ; low half of nil address
            (%get-long p 8) (ppc-lap-word (lwz 11 (+ (subprim-name->offset '.SPeabi-callback) 
                                                     (- ppc::fulltag-nil) (* -512 4)) 12))
            (%get-long p 12) (ppc-lap-word (mtctr 11))
            (%get-long p 16) (logior index (ppc-lap-word (li 11 ??)))   ; unboxed index
            (%get-long p 20) (ppc-lap-word (bctr)))
      (ff-call (%kernel-import #.ppc::kernel-import-makedataexecutable) 
               :address p 
               :unsigned-fullword 24 
               :void)
      p)))










; This is called by .SPcallback



; moved to "lib;dumplisp" as restore-pascal-functions

