;;; PPC-specific code to handle trap and uuo callbacks.

(in-package "CCL")

#+linux
(eval-when (:compile-toplevel :execute)
  (require "LINUX-RECORDS"))

;; callback here from C exception handler

(defcallback 
 %err-disp 
  (:address xp :unsigned-fullword fn-reg :unsigned-fullword pc-or-index :signed-fullword errnum :unsigned-fullword rb :signed-fullword continuable)
  (let ((fn (unless (eql fn-reg 0) (xp-gpr-lisp xp fn-reg)))
        (err-fn (if (eql continuable 0) '%err-disp-internal '%kernel-restart-internal)))
    (if (eql errnum ppc::error-stack-overflow)
      (handle-stack-overflow xp fn rb)
      (with-xp-stack-frames (xp fn frame-ptr)   ; execute body with dummy stack frame(s)
        (with-error-reentry-detection
          (let* ((rb-value (xp-gpr-lisp xp rb))
                 (res
                  (cond ((< errnum 0)
                         (%err-disp-internal errnum nil frame-ptr))
                        ((logtest errnum ppc::error-type-error)
                         (funcall err-fn 
                                  #.(car (rassoc 'type-error *kernel-simple-error-classes*))
                                  (list rb-value (logand errnum 63))
                                  frame-ptr))
                        ((eql errnum ppc::error-udf)
                         (funcall err-fn $xfunbnd (list rb-value) frame-ptr))
                        ((eql errnum ppc::error-throw-tag-missing)
                         (%err-disp-internal $xnoctag (list rb-value) frame-ptr))
                        ((eql errnum ppc::error-cant-call)
                         (funcall err-fn $xnotfun (list rb-value) frame-ptr))
                        ((eql errnum ppc::error-udf-call)
                         (return-from %err-disp
                           (handle-udf-call xp frame-ptr)))
                        ((eql errnum ppc::error-alloc-failed)
                         (%error (make-condition 
                                  'simple-storage-condition
                                  :format-string (%rsc-string $xmemfull))
                                 nil frame-ptr))
                        ((eql errnum ppc::error-memory-full)
                         (%error (make-condition 
                                  'simple-storage-condition
                                  :format-string (%rsc-string $xnomem))
                                 nil frame-ptr))
                        ((or (eql errnum ppc::error-fpu-exception-double) 
                             (eql errnum ppc::error-fpu-exception-single))
                         (let* ((code-vector (and fn fn (uvref fn 0)))
                                (instr (if code-vector 
                                         (uvref code-vector pc-or-index)
                                         (%get-long (%int-to-ptr pc-or-index)))))
                           (let* ((minor (ldb (byte 5 1) instr))
                                  (fra (ldb (byte 5 16) instr))
                                  (frb (ldb (byte 5 11) instr))
                                  (frc (ldb (byte 5 6) instr)))
                             (declare (fixnum minor fra frb frc))
                             (if (= minor 12)   ; FRSP
                               (%err-disp-internal $xcoerce (list (xp-double-float xp frc) 'short-float) frame-ptr)
                               (flet ((coerce-to-op-type (double-arg)
                                        (if (eql errnum ppc::error-fpu-exception-double)
                                          double-arg
                                          (handler-case (coerce double-arg 'short-float)
                                            (error (c) (declare (ignore c)) double-arg)))))
                                 (multiple-value-bind (status control) (xp-fpscr-info xp)
                                   (%error (make-condition (fp-condition-from-fpscr status control)
                                                           :operation (fp-minor-opcode-operation minor)
                                                           :operands (list (coerce-to-op-type 
                                                                            (xp-double-float xp fra))
                                                                           (if (= minor 25)
                                                                             (coerce-to-op-type 
                                                                              (xp-double-float xp frc))
                                                                             (coerce-to-op-type 
                                                                              (xp-double-float xp frb)))))
                                           nil
                                           frame-ptr)))))))
                        ((eql errnum ppc::error-excised-function-call)
                         (%error "~s: code has been excised." (list (xp-gpr-lisp xp ppc::nfn)) frame-ptr))
                        ((eql errnum ppc::error-too-many-values)
                         (%err-disp-internal $xtoomanyvalues (list rb-value) frame-ptr))
                        (t (%error "Unknown error #~d with arg: ~d" (list errnum rb-value) frame-ptr)))))
            (setf (xp-gpr-lisp xp rb) res)        ; munge register for continuation
            ))))))



#+ppc-target
(defun handle-udf-call (xp frame-ptr)
  ;(bug (symbol-name (xp-gpr-lisp xp ppc::fname)))
  (let* ((args (xp-argument-list xp))
         (values (multiple-value-list
                  (%kernel-restart-internal
                   $xudfcall
                   (list (xp-gpr-lisp xp ppc::fname) args)
                   frame-ptr)))
         (stack-argcnt (max 0 (- (length args) 3)))
         (vsp (%i+ (xp-gpr-lisp xp ppc::vsp) stack-argcnt))
         (f #'(lambda (values) (apply #'values values))))
    (setf (xp-gpr-lisp xp ppc::vsp) vsp
          (xp-gpr-lisp xp ppc::nargs) 1
          (xp-gpr-lisp xp ppc::arg_z) values
          (xp-gpr-lisp xp ppc::nfn) f)
    (with-macptrs ((machine-state (pref xp 
                                        :sigcontext.regs
                                        )))
      (let ((pc-offset (get-field-offset :pt-regs.nip)))
        (without-interrupts
         (%set-object machine-state
                      pc-offset
                      (%i- (%misc-address-fixnum (uvref f 0))
                           1)))))))

     ; handle_uuo will bump pc by 1



(defppclapfunction %misc-address-fixnum ((misc-object arg_z))
  (check-nargs 1)
  (la arg_z ppc::misc-data-offset misc-object)
  (blr))

; rb is the register number of the stack that overflowed.
; xp & fn are passed so that we can establish error context.
(defun handle-stack-overflow (xp fn rb)
  (flet ((stack-overflow-error ()
           (with-xp-stack-frames (xp fn frame-ptr)      ; execute body with dummy stack frame(s)
             (%error
              (make-condition
               'stack-overflow-condition 
               :format-string "Stack overflow on ~a stack.~@
                               To globally increase stack space,~@
                               increase ~s"
               :format-arguments (list
                                  (if (eql rb ppc::sp)
                                    "control"
                                    (if (eql rb ppc::vsp)
                                      "value"
                                      (if (eql rb ppc::tsp)
                                        "temp"
                                        "unknown")))
                                  '*minimum-stack-overflow-size*))
              nil frame-ptr))))
    (declare (dynamic-extent #'stack-overflow-error))
    (if (not (or (eql rb ppc::vsp)
                 (eql rb ppc::tsp)
                 (and (eql rb ppc::sp)
                      (let ((cs-area (%get-kernel-global 'current-cs)))
                        (> (%current-frame-ptr)
                           (%fixnum-ref cs-area ppc::area.softlimit))))))
      (stack-overflow-error)
      (unless *allow-stack-overflows*       ; startup
        (let* ((sg *current-stack-group*)
               (total-used (%stack-group-useable-size sg))
               (maxsize (sg.maxsize sg))
               (total-allowed (max maxsize *minimum-stack-overflow-size*))
               (delta (- total-used total-allowed))
               (diff (max (* 16 1024) (* 2 delta)))
               (continued? nil))
          (when (> delta 0)
            (unwind-protect
              (progn
                (setf (sg.maxsize sg) (+ total-allowed diff))
                (restart-case (stack-overflow-error)
                  (continue ()
                            :report (lambda (stream) 
                                      (format stream "Continue with a larger stack"))
                            (setq continued? t)
                            nil)))
              (unless continued?
                (setf (sg.maxsize sg) maxsize)))))))))
