;;;-*-Mode: LISP; Package: CCL -*-

;;; 08/31/98 gb   pretty much all new for eabi/linux.

(in-package "CCL")

(eval-when (:compile-toplevel :execute)
  (defconstant seconds-in-week (* 60 60 24 7))
  (defconstant weeks-offset 2145)
  (defconstant seconds-offset 432000)
  (defconstant minutes-per-day (* 24 60))
  (defconstant quarter-days-per-year (1+ (* 365 4)))
  (defconstant quarter-days-per-century 146097)
  (defconstant november-17-1858 678882)
  (defconstant weekday-november-17-1858 2)
)

(defun gctime ()
  (%stack-block ((copy (* 8 5)))
    (external-call "memmove" 
		   :address copy 
		   :address *total-gc-microseconds*
		   :unsigned-fullword (* 8 5)
		   :address)
    (values
     (timeval->milliseconds copy)
     (timeval->milliseconds (%incf-ptr copy 8))
     (timeval->milliseconds (%incf-ptr copy 8))
     (timeval->milliseconds (%incf-ptr copy 8))
     (timeval->milliseconds (%incf-ptr copy 8)))))

(defun get-universal-time ()
  (rlet ((tv :timeval))
    (external-call "gettimeofday" 
		   :address tv
		   :address (%null-ptr)
		   :signed-fullword)
    (+ (pref tv :timeval.tv_sec) unix-to-universal-time)))

(defun get-timezone (time)
  (rlet ((when :long))
    (setf (%get-long when) time)
    (without-interrupts			;reentrancy ?
     (with-macptrs ((ltm (external-call "localtime" :address when :address)))
       (values (floor (pref ltm :tm.__tm_gmtoff) -60)
	       (not (zerop (pref ltm :tm.tm_isdst))))))))

(defun decode-universal-time (universal-time &optional time-zone)
  "Converts a universal-time to decoded time format returning the following
   nine values: second, minute, hour, date, month, year, day of week (0 =
   Monday), T (daylight savings time) or NIL (standard time), and timezone.
   Completely ignores daylight-savings-time when time-zone is supplied."
  (multiple-value-bind (weeks secs)
		       (truncate (+ universal-time seconds-offset)
				 seconds-in-week)
    (let* ((weeks (+ weeks weeks-offset))
	   (second NIL)
	   (minute NIL)
	   (hour NIL)
	   (date NIL)
	   (month NIL)
	   (year NIL)
	   (day NIL)
	   (daylight NIL)
	   (timezone (if (null time-zone)
			 (multiple-value-bind
			     (minwest dst)
			     (get-timezone (- universal-time
					      unix-to-universal-time))
			   (setf daylight dst)
			   minwest)
			 (* time-zone 60))))
      (declare (fixnum timezone))
      (multiple-value-bind (t1 seconds) (truncate secs 60)
	(setq second seconds)
	(setq t1 (- t1 timezone))
	(let* ((tday (if (< t1 0)
			 (1- (truncate (1+ t1) minutes-per-day))
			 (truncate t1 minutes-per-day))))
	  (multiple-value-setq (hour minute)
	    (truncate (- t1 (* tday minutes-per-day)) 60))
	  (let* ((t2 (1- (* (+ (* weeks 7) tday november-17-1858) 4)))
		 (tcent (truncate t2 quarter-days-per-century)))
	    (setq t2 (mod t2 quarter-days-per-century))
	    (setq t2 (+ (- t2 (mod t2 4)) 3))
	    (setq year (+ (* tcent 100) (truncate t2 quarter-days-per-year)))
	    (let ((days-since-mar0 (1+ (truncate (mod t2 quarter-days-per-year)
						 4))))
	      (setq day (mod (+ tday weekday-november-17-1858) 7))
	      (let ((t3 (+ (* days-since-mar0 5) 456)))
		(cond ((>= t3 1989)
		       (setq t3 (- t3 1836))
		       (setq year (1+ year))))
		(multiple-value-setq (month t3) (truncate t3 153))
		(setq date (1+ (truncate t3 5))))))))
      (values second minute hour date month year day
	      daylight
	      (if daylight
		  (1+ (/ timezone 60))
		  (/ timezone 60))))))

(defun get-decoded-time ()
  (decode-universal-time (get-universal-time)))

(defun leap-years-before (year)
  (let ((years (- year 1901)))
    (+ (- (truncate years 4)
	  (truncate years 100))
       (truncate (+ years 300) 400))))

(defvar *days-before-month*
  (let* ((results (list nil)))
    (let ((sum 0))
      (dolist (days-per-month '(31 28 31 30 31 30 31 31 30 31 30 31))
	(push sum results)
	(incf sum days-per-month)))
    (coerce (nreverse results) 'vector)))

(defun encode-universal-time (second minute hour date month year
				     &optional time-zone)
  "The time values specified in decoded format are converted to 
   universal time, which is returned."
  (declare (type (mod 60) second)
	   (type (mod 60) minute)
	   (type (mod 24) hour)
	   (type (integer 1 31) date)
	   (type (integer 1 12) month)
	   (type (or (integer 0 99) (integer 1900)) year)
	   (type (or null rational) time-zone))
  (let* ((days (+ (1- date)
		  (aref *days-before-month* month)
		  (if (> month 2)
		      (leap-years-before (1+ year))
		      (leap-years-before year))
		  (* (- year 1900) 365)))
	 (hours (+ hour (* days 24))))
    (if time-zone
	(+ second (* (+ minute (* (+ hours time-zone) 60)) 60))
	(let* ((minwest-guess
		(nth-value 1
			   (get-timezone (- (* hours 60 60)
					    unix-to-universal-time))))
	       (guess (+ minute (* hours 60) minwest-guess))
	       (minwest
		(nth-value 1
			   (get-timezone (- (* guess 60)
					    unix-to-universal-time)))))
	  (+ second (* (+ guess (- minwest minwest-guess)) 60))))))

(defloadvar *ticks-per-second*
    (case (host-platform)
      (:linux 50)
      (:solaris (with-cstrs ((n "sysconf"))
                  (let* ((addr (ash
                                (ff-call
                                 (%kernel-import arch::kernel-import-dlsym)
                                 :address *rtld-next*
                                 :address n
                                 :signed-fullword)
                                -2)))
                    (ff-call addr :unsigned-fullword 3 :unsigned-fullword))))
      (:vxworks (with-cstrs ((n "sysClkRateGet"))
                  (let* ((addr (ash
                                (ff-call
                                 (%kernel-import arch::kernel-import-dlsym)
                                 :address *rtld-next*
                                 :address n
                                 :unsigned-fullword)
                                -2)))
                    (ff-call addr :unsigned-fullword))))
      (t 60)))

(defun sleep (seconds)
  (when (minusp seconds) (report-bad-arg seconds '(number 0 *)))
  (%proxy-wait-timeout (ceiling (* seconds 1000000))))



