;;;-*- Mode: Lisp; Package: CCL -*-

;;; debug-ot.lisp
;;; Copyright 1996 Digitool, Inc. 

;;; Code to help debug OpenTransport.lisp

(in-package :ccl)

(defclass ot-debug-window (fred-window)
  ((ot-stream :initarg :ot-stream :accessor ot-stream)
   (ot-direction :initform nil :accessor ot-direction)))

(defvar *ot-debug-window-hash*
  (make-hash-table :test 'eq :weak :key))

(defvar *ot-debugging-p* nil)

(defun get-ot-window (stream &optional (create t))
  (or (let ((w (gethash stream *ot-debug-window-hash*)))
        (and w (wptr w) w))
      (and create
           (let ((w (make-instance 'ot-debug-window
                      :ot-stream stream
                      :scratch-p t)))
             (format w "~s~%~%" stream)
             (setf (gethash stream *ot-debug-window-hash*) w)))))

(defun show-ot-input (stream)
  (when *ot-debugging-p*
    (let ((conn (stream-io-buffer stream nil)))
      (when (and conn (not (eql 0 (io-buffer-incount conn))))
        (let ((w (get-ot-window stream))
              (buf (io-buffer-inbuf conn)))
          (unless (eq (ot-direction w) :input)
            (format w "~%~%In:~%")
            (setf (ot-direction w) :input))          
          (dotimes (i (io-buffer-incount conn))
            (tyo (code-char (%get-byte buf i)) w))
          (force-output w))))))

(advise ot-conn-next-buffer
        (show-ot-input (io-buffer-stream (car arglist)))
        :when :after
        :name *ot-debugging-p*)

(defun show-ot-output (stream)
  (when *ot-debugging-p*
    (let* ((conn (stream-io-buffer stream nil))
           (count (%ptr-difference (io-buffer-outptr conn)
                                   (io-buffer-outbuf conn))))
      (when (and conn (not (eql 0 count)))
        (let ((w (get-ot-window stream))
              (buf (io-buffer-outbuf conn)))
          (unless (eq (ot-direction w) :output)
            (format w "~%~%Out:~%")
            (setf (ot-direction w) :output))
          (dotimes (i count)
            (tyo (code-char (%get-byte buf i)) w))
          (force-output w))))))

(advise ot-conn-flushbuf
        (show-ot-output (car arglist))
        :when :before
        :name *ot-debugging-p*)
