;;;-*- Mode: Lisp; Package: CCL -*-

;;; Record & constant defs for PPC-LINUX


(defrecord pt-regs
  (gpr (:array :unsigned-long 32))
  (nip :unsigned-long)
  (msr :unsigned-long)
  (orig-gpr3 :unsigned-long)            ; used for restarting system calls
  (ctr :unsigned-long)
  (link :unsigned-long)
  (xer :unsigned-long)
  (ccr :unsigned-long)
  (mq :unsigned-long)
  (trap :unsigned-long)
  (dar :unsigned-long)
  (dsisr :unsigned-long)
  (result :unsigned-long)
  (pad (:array :unsigned-long 5))
  (fpr (:array :double-float 32))
  (pad2 :unsigned-long)
  (fpscr :unsigned-long))

(defrecord sigcontext
  (_unused (:array :unsigned-long 4))
  (signal :signed-long)
  (handler :unsigned-long)
  (oldmask :unsigned-long)
  (regs (:pointer :pt-regs)))


(defrecord stat
  (st-dev-high :unsigned-long)
  (st-dev :unsigned-long)
  (__pad1 :unsigned-long)
  (st-ino :unsigned-long)
  (st-mode :unsigned-long)
  (st-nlink :unsigned-long)
  (st-uid :unsigned-long)
  (st-gid :unsigned-long)
  (st-rdev-high :unsigned-long)
  (st-rdev :unsigned-long)
  (__pad2 :unsigned-long)
  (st-size :long)
  (st-blksize :unsigned-long)
  (st-blocks :long)
  (st-atime :unsigned-long)             ; time of last access
  (__unused1 :unsigned-long)
  (st-mtime :unsigned-long)             ; time of last modification
  (__unused2 :unsigned-long)
  (st-ctime :unsigned-long)             ; time of last status change
  (__unused3 :unsigned-long)
  (__unused4 :unsigned-long)
  (__unused5 :unsigned-long))

;; "stat" & friends are implemented in terms of deeper, darker things that
;; need to know what version of the stat buffer they're talking about.

(defconstant __STAT_VER      3)   ; Oh.

(defconstant __S_IFMT        #o0170000); These bits determine file type.  

; File types.  
(defconstant __S_IFDIR       #o0040000); Directory.  
(defconstant __S_IFCHR       #o0020000); Character device.  
(defconstant __S_IFBLK       #o0060000); Block device.  
(defconstant __S_IFREG       #o0100000); Regular file.  
(defconstant __S_IFIFO       #o0010000); FIFO.  

; These don't actually exist on System V, but having them doesn't hurt.  
(defconstant __S_IFLNK       #o0120000); Symbolic link.  
(defconstant __S_IFSOCK      #o0140000); Socket.  

; Protection bits.  

(defconstant __S_ISUID       #o04000)  ; Set user ID on execution.  
(defconstant __S_ISGID       #o02000)  ; Set group ID on execution.  
(defconstant __S_ISVTX       #o01000)  ; Save swapped text after use (sticky).  
(defconstant __S_IREAD       #o0400)   ; Read by owner.  
(defconstant __S_IWRITE      #o0200)   ; Write by owner.  
(defconstant __S_IEXEC       #o0100)   ; Execute by owner.  

; fcntl/open modes & flags:
(defconstant O_RDONLY 0)
(defconstant O_WRONLY 1)
(defconstant O_RDWR 2)
(defconstant linux_O_CREAT #o100)
(defconstant linux_O_EXCL #o200)
(defconstant linux_O_NOCTTY #o0400)
(defconstant linux_O_TRUNC #o1000)
(defconstant linux_O_APPEND #o2000)
(defconstant linux_O_NONBLOCK #o4000)


(defmacro external (name)
  `(load-eep ,name))

(defmacro external-call (name &rest args)
  `(ff-call (eep.address (load-time-value (external ,name))) ,@args))

(make-built-in-class 'external-entry-point *istruct-class*)

(defmethod make-load-form ((eep external-entry-point))
  `(load-eep ,(eep.name eep)))

(defconstant O_CREAT #x200)
(defconstant O_EXCL #x800)

(defrecord timeval
  (tv_sec :unsigned-long)
  (tv_usec :unsigned-long))

(defrecord timezone
  (tz_minuteswest :signed-long)
  (tz_dsttime :unsigned-long))

(defrecord tm
  (tm_sec :unsigned-long)
  (tm_min :unsigned-long)
  (tm_hour :unsigned-long)
  (tm_mday :unsigned-long)
  (tm_mon :unsigned-long)
  (tm_year :unsigned-long)
  (tm_wday :unsigned-long)
  (tm_yday :unsigned-long)
  (tm_isdst :unsigned-long)
  (__tm_gmtoff :signed-long))


(defrecord rusage
  (ru_utime :timeval)
  (ru_stime :timeval)
  (ru_maxrss :long)
  (ru_ixrss :long)      ; integral shared memory size 
  (ru_idrss :long)      ; integral unshared data size 
  (ru_isrss :long)      ; integral unshared stack size 
  (ru_minflt :long)          ; page reclaims 
  (ru_majflt :long)          ; page faults 
  (ru_nswap :long)      ; swaps 
  (ru_inblock :long)         ; block input operations 
  (ru_oublock :long)         ; block output operations 
  (ru_msgsnd :long)          ; messages sent 
  (ru_msgrcv :long)          ; messages received 
  (ru_nsignals :long)        ; signals received 
  (ru_nvcsw :long)      ; voluntary context switches 
  (ru_nivcsw :long)          ; involuntary context switches 
)

(defconstant RUSAGE-SELF 0)

(defrecord termios
  (c_iflag :unsigned-long)
  (c_oflag :unsigned-long)
  (c_cflag :unsigned-long)
  (c_lflag :unsigned-long)
  (c_cc (:array :unsigned-byte 19))
  (c_line :unsigned-byte)
  (c_ispeed :unsigned-long)
  (c_ospeed :unsigned-long))

(defrecord ptaskstate
  (nexttick unsigned-long)
  (interval unsigned-long)
  (private pointer)
  (flags unsigned-integer))

(defconstant kNoThreadID 0)
(defconstant kCurrentThreadID 1)
(defconstant kApplicationThreadID 2)

(defconstant kCooperativeThread (ash 1 0))
(defconstant kPreemptiveThread (ash 1 1))

(defconstant kNewSuspend (ash 1 0))
(defconstant kUsePremadeThread (ash 1 1))
(defconstant kCreateIfNeeded (ash 1 2))
(defconstant kFPUNotNeeded (ash 1 3))
(defconstant kExactMatchThread (ash 1 4))

; This matches the (define-storage-layout area ...) form in "ccl:compiler;ppc;ppc-arch.lisp"
(defrecord gc-area
  (pred (:pointer :gc-area))
  (succ (:pointer :gc-area))
  (low :ptr)
  (high :ptr)
  (active :ptr)
  (softlimit :ptr)
  (hardlimit :ptr)
  (code :long)
  (markbits :ptr)
  (ndwords :long)
  (older (:pointer :gc-area))
  (younger (:pointer :gc-area))
  (h :ptr)
  (softprot (:pointer :protected-area))
  (hardprot (:pointer :protected-area))
  (owner :long)
  (refbits :ptr)
  (threshold :ptr)
  (gc-count :long))

(defrecord protected-area
  (next (:pointer protected-area))
  (start :ptr)                          ; first byte (page-aligned) that might be protected
  (end :ptr)                            ; last byte (page-aligned) that could be protected
  (nprot :unsigned-long)                ; Might be 0
  (protsize :unsigned-long)             ; number of bytes to protect
  (why :long))

; This matches the xframe-list struct definition in "ccl:pmcl;constants.h"
(defrecord xframe-list
  (this (:pointer :ExceptionInformation))
  (prev (:pointer :xframe-list)))

(defrecord dirent
  (d_ino :long)
  (d_offset :unsigned-long)
  (d_reclen :unsigned-short)
  (d_type :unsigned-byte)
  (d_name (:array :unsigned-byte 256)))

(defrecord passwd
  (pw_name :ptr)
  (pw_passwd :ptr)
  (pw_uid :unsigned-long)
  (pw_gid :unsigned-long)
  (pw_gecons :ptr)
  (pw_dir :ptr)
  (pw_shell :ptr))

(defconstant FIONREAD #x4004667f)
