
#include "lisp.h"
#include "lisp_globals.h"
#include "gc.h"
#include "area.h"
#include <stdlib.h>
#include <string.h>
#include "lisp-exceptions.h"
#include "loader.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#ifdef LINUX
#include <mcheck.h>
#include <dirent.h>
#include <dlfcn.h>
#include <sys/time.h>
#endif
#ifdef VXWORKS
#include <limits.h>
#include <sys/stat.h>
#include <vxWorks.h>
#include <time.h>
#include <nfsLib.h>
#include <ioLib.h>
#include <taskLib.h>
#include <arch/ppc/vxPpcLib.h>
#endif
#include <ctype.h>
#ifndef VXWORKS
#include <sys/select.h>
#endif

/*
 04/17/97 bill  Gary's fix to metering_proc to make it less likely to call
                register_codevector_contains_pc on something that will cause a bug error.
 ---- 4.1b2
 01/29/97 gb    anal-retentive extern decls for gcc.
 ---- 4.0
 09/07/96 gb    try to improve the fragLibNotFound error message.
 ---- 4.0b1
 08/01/96 gb    load_library_or_die: pass plibname vice libname to Fatal().
 07/25/96 bill  slightly de-uglify the message generated by report_library_load_failure
                by adding a few newlines.
 07/16/96 gb    allocate_lisp_stack tries application heap if MF heap fails.
                load_library_or_die & related changes.
 06/17/96 bill  allocate_lisp_stack gives you exactly as much useable space
                as you ask for. The handle will have up to 4K of wasted space
		and there will be a small unused space after the block.
 06/16/96 bill  allocate_vstack & allocate_tstack allocate 4096 bytes of soft area
                and 0 bytes of hard area.
                allocate_lisp_stack_area initializes soft_area & hard_area to NULL
                so that it will work correctly if its caller passes 0 for softsize or hardsize.
                main allocates the initial tsp & vsp stack with a useable size of
                STACK_SEGMENT_SIZE instead of 1<<16.
 06/26/96 gb    look for tagged *INTERRUPT-LEVEL* symbol in cmain_vbl.  Create ephemeral
                areas.  Fatal errors from loader propagate here.
 06/14/96 slh   continuable_error_dialog: no exit if no dialog
 06/05/96 slh   splash screen
 05/12/96 gb    startup changes for new (non-PEF data) loader.
 ---- 3.9
 04/26/96 gb    account for heap_spare when allocating global_reloctab.
 04/15/96 gb    lisp_heap_size_from_LSIZ: give lisp 2% more (since no static refbits and
                markbits allocated from dynamic heap handle.)
 ---- 3.9f1c2
 04/11/96 gb    allocate_lisp_heap: adjust a->high & a->ndwords when reserving heap_spare.
 04/10/96 gb    new heap allocation stuff.
 04/09/96 gb    use a dialog to warn about missing exception support.
                call set_fpscr to initialize main thread's fpscr.
 04/08/96 bill  initialize EXCEPTION_SAVED_REGISTERS
 03/20/96 gb    second arg to start_lisp.
 03/21/96 bill  lisp_heap_size_from_LSIZ_resource reserves 5% for the mark bits.
                This makes it work leave sufficient Mac heap for all application sizes.
 03/18/96 bill  from gb: "#pragma options align=mac68k" around LSIZ_resource struct definition
 03/18/96 gb    conditionalize out lisp_heap_size_from_LSIZ_resource.  (seemed
                to cause GC failure; not sure why. 
 03/15/96 bill  Use lisp_heap_size_from_LSIZ_resource to determine the
                size of the lisp heap.
 03/15/96 bill  from gb: switch AREA_STATIC & AREA_STATICLIB in
                new new_area calls in set_nil_and_start
 03/01/96 bill  Update comment for cmain_vbl
 03/01/96 gb    protection racket; metering.
 02/20/96 gb    new scheme for all_areas; stop at tb when fixing
                code vectors.
 02/28/96 bill  install_lisp_growzone() just before start_lisp(...)
 02/18/96 bill  allocate_lisp_stack handles NULL value from new_protected_area
 01/29/96 gb    keep track of protected_areas in areas.
 01/26/96 gb    initialize LEXPR_RETURN, LEXPR_RETURN1V
 01/22/96 gb    stack allocation changes.
 01/16/96 gb    allocate_lisp_heap returns handle.
 12/27/95 gb    vbl descriptor in sys heap.
 12/22/95 gb    no MoveHHi (VM thrashes); init READONLY_AREA
 12/13/95 gb    heap allocation changes; main() returns 0.
 11/23/95 bill  start_vbl, cmain_vbl
 11/19/95 gb    no mo SIOW
 10/26/95 slh   set HEAP_START, HEAP_END globals
*/

/* This is kept outside of the kernel_globals structure
   'cause it's referenced so often. */

LispObj lisp_nil = (LispObj) 0;


/* These are all "persistent" : they're initialized when
   subprims are first loaded and should never change. */
extern LispObj (*start_lisp)(LispObj, LispObj);
extern LispObj subprims_base;
extern LispObj ret1valn;
extern LispObj lexpr_return;
extern LispObj lexpr_return1v;
LispObj real_subprims_base = 0;
LispObj text_start = 0;

/* A pointer to some of the kernel's own data; also persistent. */

extern LispObj import_ptrs_base;

void remap_subprims(void);

unsigned
align_to_power_of_2(unsigned n, unsigned power)
{
  unsigned align = (1<<power) -1;

  return (n+align) & ~align;
}

void
setup_target_subprims(void)
{
  LispObj w, *jtabp = (LispObj *) real_subprims_base, disp, dest;
  int target = SUBPRIMS_TARGET_0, i;
  
  for (i = 0; i < 256; i++, target++, jtabp++) {
    w = *jtabp;
    if ((w >> 26) != 18) break;  /* 18 = b[l][a] opcode */
    disp = w & (((1<<26)-1) & ~3);
    dest = (LispObj) jtabp+disp;
    lisp_global(target) = dest;
  }
}


void
relocate_code_vector(LispObj *entry,/* first word past header */
                     unsigned n, /* word count */
                     LispObj bias) /* subprims_base */
{
  LispObj w;
  
  for (;n--;entry++) {
    w = *entry;
    if (w == 0) { 
      break;
    }
    if ((w & BA_MASK) == BA_VAL) {
      *entry = (w + bias);      /* Add it, don't OR it. */
    }
  }
}


void
MakeDataExecutable(void *, unsigned);

void
relocate_dynamic_subprims(LispObj *p, LispObj *q)
{
  LispObj base = lisp_global(SUBPRIMS_BASE);
  
  void * cache_start = (void *) p;
  unsigned ncacheflush = (unsigned) q - (unsigned) p;
  LispObj header;
  
  int nelements, tag, subtag;
  
  while (p < q) {
    header = *p;
    tag = fulltag_of(header);
    if ((tag != fulltag_nodeheader) && (tag != fulltag_immheader)) {
      p += 2;
    } else {
      nelements = header_element_count(header);
      if (tag == fulltag_nodeheader) {
        nelements = (nelements+1+1) & ~1;
        p += nelements;
      } else {
        subtag = header_subtag(header);
        if (subtag == subtag_code_vector) {
          relocate_code_vector(p+1, nelements, base);
        }
        p = (LispObj *) skip_over_ivector((unsigned)p, header);
      }
    }
  }
  MakeDataExecutable(cache_start, ncacheflush);  
}
      


/* This should write-protect the bottom of the stack.
   Doing so reliably involves ensuring that everything's unprotected on exit.
*/

static BytePtr
allocate_lisp_stack(unsigned useable,
                    unsigned softsize,
                    unsigned hardsize,
                    lisp_protection_kind softkind,
                    lisp_protection_kind hardkind,
                    Ptr *h_p,
                    BytePtr *base_p,
                    protected_area_ptr *softp,
                    protected_area_ptr *hardp)
{
  unsigned size = useable+softsize+hardsize;
  unsigned overhead;
  BytePtr base, softlimit, hardlimit;
  OSErr err;
  Ptr h = allocate(size+4095);
  protected_area_ptr hprotp = NULL, sprotp;

  if (h == NULL) {
    return NULL;
  }
  if (h_p) *h_p = h;
  base = (BytePtr) align_to_power_of_2((unsigned) h, 12);
  hardlimit = (BytePtr) (base+hardsize);
  softlimit = hardlimit+softsize;

  overhead = (base - (BytePtr) h);
  if (hardsize) {
    hprotp = new_protected_area((BytePtr)base,hardlimit,hardkind, hardsize, false);
    if (hprotp == NULL) {
      if (base_p) *base_p = NULL;
      if (h_p) *h_p = NULL;
      deallocate(h);
      return NULL;
    }
    if (hardp) *hardp = hprotp;
  }
  if (softsize) {
    sprotp = new_protected_area(hardlimit,softlimit, softkind, softsize, true);
    if (sprotp == NULL) {
      if (base_p) *base_p = NULL;
      if (h_p) *h_p = NULL;
      if (hardp) *hardp = NULL;
      if (hprotp) delete_protected_area(hprotp);
      deallocate(h);
      return NULL;
    }
    if (softp) *softp = sprotp;
  }
  if (base_p) *base_p = base;
  return (BytePtr) ((unsigned)(base+size));
}

/* This'll allocate a tstack or a vstack, but the thread
   mangler won't let us allocate or reliably protect
   a control stack.
*/
area *
allocate_lisp_stack_area(area_code stack_type,
                         unsigned useable, 
                         unsigned softsize, 
                         unsigned hardsize, 
                         lisp_protection_kind softkind, 
                         lisp_protection_kind hardkind)

{
  BytePtr base, bottom;
  Ptr h;
  area *a = NULL;
  protected_area_ptr soft_area=NULL, hard_area=NULL;

  bottom = allocate_lisp_stack(useable, 
                               softsize, 
                               hardsize, 
                               softkind, 
                               hardkind, 
                               &h, 
                               &base,
                               &soft_area, 
                               &hard_area);

  if (bottom) {
    a = new_area(base, bottom, stack_type);
    a->hardlimit = base+hardsize;
    a->softlimit = base+hardsize+softsize;
    a->h = h;
    a->softprot = soft_area;
    a->hardprot = hard_area;
    add_area(a);
  }
  return a;
}

area*
register_cstack(BytePtr bottom, unsigned size)
{
  BytePtr lowlimit = (BytePtr) (((((unsigned)bottom)-size)+4095)&~4095);
  area *a = new_area((BytePtr) bottom-size, bottom, AREA_CSTACK);

  a->hardlimit = lowlimit+4096;
  a->softlimit = lowlimit+8192;
  add_area(a);
  return a;
}
  
area*
allocate_vstack(unsigned useable)
{
  return allocate_lisp_stack_area(AREA_VSTACK, 
                                  useable, 
                                  4096,
                                  0,
                                  kVSPsoftguard,
                                  kVSPhardguard);
}

area *
allocate_tstack(unsigned useable)
{
  return allocate_lisp_stack_area(AREA_TSTACK, 
                                  useable, 
                                  4096,
                                  0,
                                  kTSPsoftguard,
                                  kTSPhardguard);
}

typedef struct LSIZ_resource LSIZ_resource, *LSIZ_ptr;

#pragma options align=mac68k
struct LSIZ_resource {
  long mac_heap_minimum;
  long mac_heap_maximum;
  short mac_heap_percentage;
  long low_memory_threshold;
  long copying_gc_threshold;
  long stack_minimum;
  long stack_maximum;
  short stack_percentage;
};
#pragma options align=reset

struct LSIZ_resource fake_LSIZ = {
  100 << 10,
  400 << 10,
  5,
  (64 << 10),
  0x7fffffff,
  (32 << 10),
  (180 << 10),
  6
};

/* It's hard to believe that max & min don't exist already */
unsigned unsigned_min(unsigned x, unsigned y)
{
  if (x <= y) {
    return x;
  } else {
    return y;
  }
}

unsigned unsigned_max(unsigned x, unsigned y)
{
  if (x >= y) {
    return x;
  } else {
    return y;
  }
}

unsigned
heap_spare_size = 65536;        /* size to hold in reserve */
unsigned
heap_spare = 0;                 /* size currently held in reserve */

unsigned
lisp_heap_size_from_LSIZ_resource (unsigned max_block)
{
  LSIZ_ptr lsiz = (LSIZ_ptr) &fake_LSIZ;

  if (lsiz == NULL) {
    return (max_block*3)/4;
  } else {
    unsigned free_bytes = max_block;
    unsigned desired_macheap_size = 
             (unsigned) (((double)free_bytes) * ((double) lsiz->mac_heap_percentage) / 100.0);
    unsigned real_macheap_size = 
             unsigned_min(lsiz->mac_heap_maximum,
                          unsigned_max(lsiz->mac_heap_minimum, desired_macheap_size));
    unsigned result = free_bytes - real_macheap_size - 4095 - heap_segment_size;
    unsigned spare = (lsiz->low_memory_threshold);
    /* 3% for mark bits and other random stuff.  This was 5%, but we're
       not taking the markbits out of a seperate heap block anymore.
     */
    result = (unsigned) (((double) result) / 1.03);
    if (spare) {
      heap_spare_size = align_to_power_of_2(spare, log2_heap_segment_size);
    }
    return result;
  }
}

int
lisp_partition_exists = 0;

Ptr
lisp_heap_partition = NULL;

void
create_kernel_globals()
{
  KG = (kernel_globals *)allocate(sizeof(kernel_globals));
  if (KG != NULL) {
    bzero((void *)KG, sizeof(kernel_globals));
    KG->cache_block_size = 32;	/* CPU-specific */
  } else {
    fprintf(stdout, "Can't create kernel globals structure\n");
    fflush(stdout);
    lisp_exit(-1);
  }
}
Ptr
create_memory_partition(unsigned totalsize)
{
  lisp_heap_partition = malloc(totalsize);
  if (lisp_heap_partition != NULL) {
    dl_malloc_init(lisp_heap_partition, totalsize);
    lisp_partition_exists = 1;
  } else {
    lisp_exit(-1);
  }
  return lisp_heap_partition;
}


int 
lisp_heap_size = (8<<20);

area *
allocate_lisp_heap(unsigned size)
{
  OSErr err;
  Ptr h;
  unsigned base, totalsize;
  BytePtr suffix, end, lastbyte, start, protstart;
  area *a;
  protected_area_ptr softp, hardp;
  bitvector markbits;

  if (size == 0) {
    size = (lisp_heap_size * 3)/8;
  }
  /* size = lisp_heap_size_from_LSIZ_resource(size ? size : lisp_heap_size); */

  totalsize = 4095+align_to_power_of_2(size+heap_segment_size,log2_heap_segment_size);

  h = allocate(totalsize);
  if (h == NULL) {
    return NULL;                /* caller must check for this! */
  }
  base = (unsigned) h;
  lastbyte = (BytePtr) (base+totalsize);
  end = (BytePtr) ((int)lastbyte  & ~4095);
  suffix = end-heap_segment_size;
  hardp = new_protected_area(suffix,end,kHEAPhard, heap_segment_size, false);
  start = (BytePtr) (align_to_power_of_2(base,3));
  /* Allocate mark bits here.  They need to be 1/64 the size of the maximum useable
     area of the heap, or 1/65 its current size.  Round up to a word boundary, add
     three words (for egc, eventually), and align "start" to an 8-byte boundary. */
  markbits = (bitvector) start;
  start += align_to_power_of_2(12+((suffix-start)/65), 3);
  protstart = (BytePtr) (suffix - (((suffix-start)>>log2_heap_segment_size)<<log2_heap_segment_size));
  softp = new_protected_area(protstart,suffix,kHEAPsoft, heap_segment_size, true);
  memset(start,0,protstart-start);
  a = new_area(start, suffix, AREA_DYNAMIC);
  add_area(a);
  a->markbits = markbits;
  a->h = h;
  a->softprot = softp;
  a->hardprot = hardp;
  a->hardlimit = hardp->start;
  /* Try to reserve some spare bytes in the hard protected area 
     (by taking them out of the soft area.) */
  if (((softp->end - softp->start) - softp->nprot) > heap_spare_size) {
    softp->end -= heap_spare_size;
    hardp->start -= heap_spare_size;
    a->high -= heap_spare_size;
    a->ndwords -= (heap_spare_size >> 3);
    heap_spare = heap_spare_size;
  }
  return a;
}


/* *interrupt-level* is >= 0 when interrupts are enabled and < 0
   during without-interrupts. Normally, it is 0. When this VBL
   goes off, it sets it to 1 if it's 0, or if it's negative,
   walks up the special binding list looking for a previous
   value of 0 to set to 1. 
   Inside Mac says that a VBL will not be run while it's process
   is in the background unless the handler is in the system
   heap. I don't know if this is still true on the PPC.
   The handler is also supposed to be wired into memory, though Gary
   says that VBL's aren't run unless it's OK to swap. */
void
preemption_handler(int level)
{
  struct lispsymbol *interrupt_level_sym = &nrs_INTERRUPT_LEVEL;
  int interrupt_level = (int) interrupt_level_sym->vcell;
  LispObj tagged_il_sym = ((LispObj) interrupt_level_sym) + fulltag_misc;

  lisp_global(TICKS) += (1<<fixnumshift); /* Screw: handle overflow ? */

  if (interrupt_level == 0) {
    nrs_INTERRUPT_LEVEL.vcell = level<<fixnumshift;
  }
  else if ((tag_of(interrupt_level) == tag_fixnum) && /* This test may not be necessary */
           (interrupt_level < 0)) {
    struct special_binding *b;
    LispObj b_value;
    for (b = (special_binding *) lisp_global(DB_LINK); b != 0; b = b->link) {
      if (((b->sym) == (lispsymbol *) tagged_il_sym)) {
        if ((b_value = (b->value)) == 0) {
          b->value = level<<fixnumshift;
          break;
        }
        else if (b_value > 0) {
          break;
        }
      }
    }
  }
}

#ifdef LINUX
#define WHAT_ITIMER ITIMER_REAL
#define WHAT_TIMER_SIGNAL SIGALRM
#endif


void
alarm_handler (int signum)
{
#ifdef LINUX
  if (signum == WHAT_TIMER_SIGNAL) {
    preemption_handler(1);
  }
#endif
  if (signum == SIGINT) {
    lisp_global(INTFLAG) = (1 << fixnumshift);
  }
}




/* Start up the VBL task that frobs *interrupt-level* to cause an interrupt. */
#ifdef LINUX
void
start_vbl()
{
  struct itimerval vbl_timer = {
    {0, 20000},
    {0, 20000}};
  signal(WHAT_TIMER_SIGNAL, alarm_handler);
  setitimer(WHAT_ITIMER, &vbl_timer, NULL);
  signal(SIGINT, alarm_handler);

}
#endif

#ifdef VXWORKS
int vxworks_timer_enabled = 1;

void
vxworks_timer_proc()
{
  while(vxworks_timer_enabled) {
    taskDelay(1);
    preemption_handler(1);
  }
  exit(0);
}

void
start_vbl()
{
  WIND_TCB *tcb = taskTcb(0);
  int priority = tcb->priority - 1;
  extern int lisp_timer_task;

  if (priority < 0) {
    priority = 0;
  }
  lisp_timer_task = taskSpawn("ppccl_timer", priority, 0, 8<<10, (FUNCPTR) vxworks_timer_proc, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10);
  signal(SIGINT, alarm_handler);
}

#endif

extern BytePtr
current_stack_pointer(void);



  
Ptr fatal_spare_ptr = NULL;

void
prepare_for_the_worst()
{
  /* I guess that CouldDialog is no more */
  /* CouldDialog(666); */
}

void
Fatal(StringPtr param0, StringPtr param1)
{

  if (fatal_spare_ptr) {
    deallocate(fatal_spare_ptr);
    fatal_spare_ptr = NULL;
  }
  fprintf(stdout, "Fatal error: %s\n%s\n", param0, param1);
  lisp_exit(-1);
}

OSErr application_load_err = noErr;
extern Boolean load_subprims(char *);
#ifdef VXWORKS
extern Boolean load_vxlow(char *);
#endif

area *
set_nil(LispObj);

#ifdef VXWORKS
extern char *
strdup(char *);
char * default_lisp_image_name = "./VXPPCCL";
#endif

char *
default_image_name(char *orig)
{
  char *copy = strdup(orig), *base = copy, *work = copy, c;
  if (copy == NULL) {
    return NULL;
  }
  while(*work) {
    if (*work++ == '/') {
      base = work;
    }
  }
  work = base;
  while (c = *work) {
    if (islower(c)) {
      *work++ = toupper(c);
    } else {
      *work++ = tolower(c);
    }
  }
  return copy;
}

/*
  Cleanup everything so that we can run some lisp image again.
  Free everything allocated in the lisp_heap_partition and
  reset the (private) malloc that we use.
  */

void
image_memory_cleanup()
{
  KG = NULL;
  free(lisp_heap_partition);
  lisp_partition_exists = 0;
  zone_malloc_reset();
  lisp_nil = (LispObj)NULL;
}

int no_sigtrap = 0;

#ifdef LINUX
main(int argc, char *argv[])
#endif
#ifdef VXWORKS
vxppcclMain()
#endif
{
#ifdef VXWORKS
  int argc = 1;
  char *argv[] = {"vxppccl", 0};
#endif
  extern  set_fpscr(unsigned);
  extern LispObj load_image(char *);
  long resp;
  BytePtr heap_start, heap_end, stack_end;
  area *a;
  char *image_name;


#ifdef LINUX
  image_name = default_image_name(argv[0]);
#endif
  if (argc > 1) {
    int i;
    for (i = 1; i < argc; i++) {
      if (!(strcmp(argv[i], "--no-sigtrap"))) {
	no_sigtrap = 1;
      } else {
	image_name = argv[i];
      }
    }
  } else {
#ifdef VXWORKS
#if 0
    ioTaskStdSet(0, 2, ioTaskStdGet(0, 1));
#endif
    image_name = default_lisp_image_name;
#else
    image_name = default_image_name(argv[0]);
#endif
  }
#ifdef VXWORKS
  vxMsrSet(vxMsrGet() & ~0x100);
#endif
  create_memory_partition(lisp_heap_size);
  create_kernel_globals();
#if 0
  fcntl(0, F_SETFL, O_NONBLOCK |  fcntl(0, F_GETFL));
#endif

  prepare_for_the_worst();

  if (!load_subprims("./lisp_subprims.o")) {
    fprintf(stdout, "Couldn't load subprims.\n");
    lisp_exit(-1);
  }

  real_subprims_base = subprims_base;
#ifdef VXWORKS
  if (!load_vxlow("./vxlow.o")) {
    fprintf(stdout, "Couldn't load vxlow module.\n");
    lisp_exit(-1);
  }
#endif
  set_nil(load_image(image_name));
  remap_subprims();
  exception_init();
  puppet_init();

  if (lisp_global(SUBPRIMS_BASE) == 0) {
    Fatal(": Couldn't load subprims library.", "");
  }
  
  if (application_load_err != noErr) {
    Fatal(": Error loading application data.", (load_error_message != NULL) ? load_error_message : (StringPtr) "Unknown.");
  }

  set_fpscr(0xD0);


  lisp_global(KERNEL_IMPORTS) = (LispObj) import_ptrs_base;
  if (lisp_global(SAVE_FREEPTR) == 0) {
    a = allocate_lisp_heap(0);
    if (a == NULL) {
      Fatal(": Couldn't allocate lisp heap; memory must be VERY tight.", "");
    }
    lisp_global(SAVE_FREEPTR) = (LispObj) (a->low);
  } else {
    a = active_dynamic_area;
  }
  heap_start = a->low;
  heap_end = a->high;

  *(double *) &(lisp_global(DOUBLE_FLOAT_ONE)) = (double) 1.0;
  lisp_global(EXCEPTION_SAVED_REGISTERS) = (LispObj) 0;

  
  global_reloctab = (LispObj *) 
    allocate(sizeof(LispObj)*((((a->ndwords)+(heap_spare>>3)+31)>>5)+1));
  
  lisp_global(HEAP_START) = (LispObj) heap_start;
  lisp_global(HEAP_END) = (LispObj) heap_end;

  lisp_global(ARGV) = (LispObj) argv;
  lisp_global(ERRNO) = (LispObj) &errno;
  lisp_global(HOST_PLATFORM) = (LispObj)
#ifdef LINUX
    1
#endif
#ifdef VXWORKS
    2
#endif
    /*We'll get a syntax error here if neither LINUX or VXWORKS is defined. */
    << fixnumshift;


  if (KG->nilreg_area != NULL) {
    BytePtr lowptr = (BytePtr) a->low;

    a = active_dynamic_area;
    /* Create these areas as AREA_STATIC, change them to AREA_DYNAMIC */
    KG->g1_area = new_area(lowptr, lowptr, AREA_STATIC);
    KG->g2_area = new_area(lowptr, lowptr, AREA_STATIC);
    KG->tenured_area = new_area(lowptr, lowptr, AREA_STATIC);
    add_area(KG->tenured_area);
    add_area(KG->g2_area);
    add_area(KG->g1_area);

    KG->g1_area->code = AREA_DYNAMIC;
    KG->g2_area->code = AREA_DYNAMIC;
    KG->tenured_area->code = AREA_DYNAMIC;

/*    a->older = g1_area; */ /* Not yet: this is what "enabling the EGC" does. */
    KG->g1_area->younger = a;
    KG->g1_area->older = KG->g2_area;
    KG->g2_area->younger = KG->g1_area;
    KG->g2_area->older = KG->tenured_area;
    KG->tenured_area->younger = KG->g2_area;
    KG->tenured_area->refbits = a->markbits;
    lisp_global(TENURED_AREA) = (LispObj)(KG->tenured_area);
    KG->g2_area->threshold = (1<<18); /* 256K */
    KG->g1_area->threshold = (1<<17); /* 128K */
    a->threshold = (1<<16);     /* 64K */
  }

  memo_base =
    (BytePtr) allocate_lisp_stack(1<<15, 1<<12, 0, kMEMOprotect, kNotProtected, NULL, NULL, NULL, NULL);
  lisp_global(SAVE_MEMO) = (LispObj) memo_base;
      
  a = allocate_vstack(STACK_SEGMENT_SIZE);
  stack_end = a->high;
  lisp_global(SAVE_VSP) = (LispObj) stack_end;
  lisp_global(CURRENT_VS) = (LispObj) a;

  a = allocate_tstack(STACK_SEGMENT_SIZE);
  stack_end = a->high;
  lisp_global(SAVE_TSP) = (LispObj) stack_end;
  lisp_global(CURRENT_TS) = (LispObj) a;

  a = register_cstack(current_stack_pointer(), StackSpace());
#ifndef MACOS
  init_mac_threads((void *)(current_stack_pointer() - StackSpace()));
#endif
  lisp_global(CURRENT_CS) = (LispObj) a;
  lisp_global(CS_OVERFLOW_LIMIT) = (LispObj) (a->softlimit);

#if 1
  start_vbl();
#endif

  start_lisp(lisp_nil, 0);
  lisp_exit(0);
}

int
lisp_exit(int status)
{
#ifdef VXWORKS
  vxworks_timer_enabled = 0;
#endif
  exception_cleanup();
  image_memory_cleanup();
  exit(status);
  return status;
}


mutable_data_section_header **next_libP = NULL;
unsigned next_libnum = 0;


area *
set_nil(LispObj r)
{
  static_header *static_headerP;
  area *root, *a, *primary;
  mutable_data_section_header *libP;
  BytePtr low, high;
  unsigned libnum = 2;

  if (lisp_nil == (LispObj)NULL) {
    root = new_area(NULL, NULL, AREA_VOID);
    if (root == NULL) {
      lisp_exit(-1);
    }

    /* The root of all evil is initially linked to itself. */
    root->succ = root->pred = root;
    KG->all_areas = root;

    lisp_nil = r;
    lisp_global(SUBPRIMS_BASE) = subprims_base;
    lisp_global(RET1VALN) = ret1valn;
    lisp_global(LEXPR_RETURN) = lexpr_return;
    lisp_global(LEXPR_RETURN1V) = lexpr_return1v;

    static_headerP = (static_header *) (&(lisp_global(STATIC_HEAP_START)));
    next_libP = &(static_headerP->next);

    low = static_headerP->low;
    high = static_headerP->high;
    
    primary = new_area(low, high, AREA_STATIC);
    add_area(primary);

    setup_target_subprims();
    relocate_dynamic_subprims((LispObj *) (low),
                              (LispObj *) (high));

    for (libP = static_headerP->next; libP; libP = libP->next) {
      BytePtr *lowP;
      
      /*      load_library_or_die(libP); */
      low = libP->mutable_low;
      a = new_area(low, libP->mutable_high, AREA_STATIC);
      a->owner = libnum;
      add_area(a);

      low = libP->immutable_low;
      lowP = (BytePtr *) low;
      a = new_area(low, libP->immutable_high, AREA_STATICLIB);
      a->owner = libnum;
      add_area(a);

      a = new_area(lowP[0], lowP[1], AREA_READONLY);
      a->owner = libnum;
      add_area(a);


      libnum++;
      next_libP = &(libP->next);
    }

    next_libnum = libnum;

    static_headerP = (static_header *) (&(lisp_global(READONLY_SECTION_START)));
    if (static_headerP->low) {
      add_area(new_area(static_headerP->low, static_headerP->high, AREA_READONLY));
    }

    lisp_global(ALL_AREAS) = (LispObj) (KG->all_areas);
    return primary;
  }
  return NULL;
}


LispObj init_code_vector = (LispObj)NULL;


OSErr
application_loader(int fd, long pos, LispObj r)
{
  area *a;
  init_code_vector = deref(r,0);
  deref(r, 0) = (LispObj)NULL;

  KG->nilreg_area = set_nil(r);


  a = allocate_lisp_heap(0);
  application_load_err = load_application_data(a, fd, pos);
  lisp_global(SAVE_FREEPTR) = (LispObj) (a->active);
  return noErr;
}


/* 
   Dynamically compiled or fasloaded code needs to be able to call
   subprims with absolute branch instructions; this means that the
   target of those instructions (the value of "subprims_base") must
   be in the low 25 bits (or the high 25) of the address space.
   If subprims_base isn't in the right place, the text section 
   might be; we can use the jump table at the start of the text
   section.  Otherwise, we'll have to copy that jump table into
   a page in the low 25 bits of the address space.

   If we don't wind up using the text section, we found something
   better, and can fix up the text section's jump table by turning
   its relative branches into absolute ones.  SAVE-APPLICATION would
   have to know that we did this and undo it.

   Whatever this does, 'real_subprims_base' is left pointing to
   the real live subprims jump table.

 */
void
remap_subprims()
{
  if (subprims_base != (subprims_base & ((1 << 25) -1))) {
    if (text_start == (text_start & ((1 << 25) - 1))) {
      subprims_base = text_start + 8;
    } else {
#ifdef LINUX
      int zero_fd = open("/dev/zero", O_RDONLY, 0666);
      char *page, *res;
      int 
	prot = PROT_READ | PROT_WRITE | PROT_EXEC,
	flags = MAP_PRIVATE | MAP_FIXED;
      if (zero_fd >= 0) {
	for (page = (char *) 4096; (unsigned) page < (1 << 25); page += 4096) {
	  if (mmap(page, 4096, prot, flags, zero_fd, 0) == page) {
	    memcpy(page, (char *) text_start+8, 4096);
	    MakeDataExecutable(page, 4096);
	    subprims_base = (LispObj) page;
	    return;
	  }
	}
	close(zero_fd);
      }
#endif /* #ifdef LINUX */
#ifdef VXWORKS
      char *page = (char *) malloc(4096);
      if (page != NULL) {
	memcpy(page, (char *) text_start+8, 4096);
	MakeDataExecutable(page, 4096);
	subprims_base = (LispObj) page;
	return;
      }
#endif
      fprintf(stdout, "Couldn't map subprims into low memory\n");
      lisp_exit(2);
    }
  } else {
    /* subprims_base is OK.  Clobber the text section's PC-relative jump
       table, replacing each "b trampolineN" with "ba realcodeN".  As noted
       above, SAVE-APPLICATION has to rewrite the jump table we clobber here. */
    LispObj *p = ((LispObj *)text_start)+2, spaddr;
    int g;

    for (g = SUBPRIMS_TARGET_0; (spaddr = lisp_global(g)) != (LispObj)NULL; g++) {
      *p++ = BA_VAL | spaddr;
    }
    MakeDataExecutable((char *)text_start, 4096);
  }
}

#ifdef VXWORKS
char *strdup(char *src)
{
  char *dest = (char *) allocate(strlen(src)+1);
  if (dest != NULL) {
    strcpy(dest, src);
  }
  return dest;
}
#endif

      
#ifndef MACOS
void
MakeDataExecutable(void *start, unsigned nbytes)
{
  extern void flush_cache_lines();
  unsigned ustart = (unsigned) start, base, end;
  
  base = (ustart) & ~(KG->cache_block_size-1);
  end = (ustart + nbytes + KG->cache_block_size - 1) & ~(KG->cache_block_size-1);
  flush_cache_lines(base, (end-base)/KG->cache_block_size, KG->cache_block_size);
}

int
StackSpace()
{
#ifdef LINUX
  return 32<<10;
#endif
#ifdef VXWORKS
  WIND_TCB *tcb = taskTcb(0);
  char here;
  return &here - tcb->pStackLimit;
#endif
}

void *
GetSharedLibrary(char *path, int mode)
{
#ifdef VXWORKS
  return NULL;
#else
  return dlopen(path, mode);
#endif
}

void *
xdlopen(char *path, int mode)
{
#ifndef VXWORKS
  return dlopen(path, mode);
#endif
#ifdef VXWORKS
  return (void *) -1;
#endif
}

void
FindSymbol(void)
{
  Bug(NULL, "FindSymbol doesn't exist.");
}

void
CountSymbols(void)
{
  Bug(NULL, "CountSymbols doesn't exist.");
}

void
GetIndSymbol(void)
{
  Bug(NULL, "GetIndSymbol doesn't exist.");
}

void
metering_control(void)
{
  Bug(NULL, "metering_control doesn't exist.");
}

void
set_nil_and_start()
{
}

int
ioctl_1(int fd, int selector, int *arg)
{
#ifdef VXWORKS
  switch(selector) {
  case 0x4004667f:
    selector = FIONREAD;
    break;
  }
#endif
  return ioctl(fd, selector, (int)arg);
}

int
xftruncate(int fd, int len)
{
#ifndef VXWORKS
  return ftruncate(fd, len);
#endif
#ifdef VXWORKS
  return -1;
#endif
}


char *
xgetcwd(char *buf, int len)
{
  char *p = getcwd(buf, len);
#ifdef VXWORKS
  if (*p != '/') {
    char *q = strchr(p, ':');
    
    if ((q != NULL) &&
	(*(++q) == '/')) {
      strcpy(p, q);
    }
  }
#endif
  return p;
}

#ifdef VXWORKS
#include <sys/stat.h>

struct linux_stat {
  int st_dev_high;
  int st_dev;
  unsigned __pad1;
  unsigned st_ino;
  unsigned st_mode;
  unsigned st_nlink;
  unsigned st_uid;
  unsigned st_gid;
  unsigned st_rdev_high;
  unsigned st_rdev;
  unsigned __pad2;
  long st_size;
  unsigned st_blksize;
  long st_blocks;
  unsigned st_atime;
  unsigned __unused1;
  unsigned st_mtime;
  unsigned __unused2;
  unsigned st_ctime;
  unsigned __unused3;
  unsigned __unused4;
  unsigned __unused5;
};

void
copy_stat_results(struct linux_stat *l, struct stat *s)
{
  l->st_dev = s->st_dev;
  l->st_ino = s->st_dev;
  l->st_mode = s->st_mode;
  l->st_nlink = s->st_nlink;
  l->st_uid = s->st_uid;
  l->st_gid = s->st_gid;
  l->st_rdev = s->st_rdev;
  l->st_size = s->st_size;
  l->st_atime = s->st_atime;
  l->st_mtime = s->st_mtime;
  l->st_ctime = s->st_ctime;
  l->st_blksize = s->st_blksize;
  l->st_blocks = s->st_blocks;
}
    
int
__xstat(unsigned long version, char *name, struct linux_stat *l)
{
  struct stat s_, *s = &s_;
  int result = stat(name, s);
  if (result == 0) {
    copy_stat_results(l,s);
  }
  return result;
}

int
__lxstat(unsigned long version, char *name, struct linux_stat *l)
{
  /* VxWorks doesn't seem to have an "lstat". */
  return __xstat(version,name,l);
}

void
report_unimplemented(char *name)
{
  char buf[128];
  sprintf(buf, "call to unimplemented function %s\n", name);
  Bug(NULL,buf);
}

int
gettimeofday(struct timespec *tm, void *tz)
{
  clock_gettime(CLOCK_REALTIME, tm);
  tm->tv_nsec /= 1000;
  return 0;
}



struct rusage
{
  /* Total amount of user time used.  */
  struct timespec ru_utime;
  /* Total amount of system time used.  */
  struct timespec ru_stime;
  /* Maximum resident set size (in kilobytes).  */
  long int ru_maxrss;
  /* Amount of sharing of text segment memory
     with other processes (kilobyte-seconds).  */
  long int ru_ixrss;
  /* Amount of data segment memory used (kilobyte-seconds).  */
  long int ru_idrss;
  /* Amount of stack memory used (kilobyte-seconds).  */
  long int ru_isrss;
  /* Number of soft page faults (i.e. those serviced by reclaiming
     a page from the list of pages awaiting reallocation.  */
  long int ru_minflt;
  /* Number of hard page faults (i.e. those that required I/O).  */
  long int ru_majflt;
  /* Number of times a process was swapped out of physical memory.  */
  long int ru_nswap;
  /* Number of input operations via the file system.  Note: This
     and `ru_oublock' do not include operations with the cache.  */
  long int ru_inblock;
  /* Number of output operations via the file system.  */
  long int ru_oublock;
  /* Number of IPC messages sent.  */
  long int ru_msgsnd;
  /* Number of IPC messages received.  */
  long int ru_msgrcv;
  /* Number of signals delivered.  */
  long int ru_nsignals;
  /* Number of voluntary context switches, i.e. because the process
     gave up the process before it had to (usually to wait for some
     resource to be available).  */
  long int ru_nvcsw;
  /* Number of involuntary context switches, i.e. a higher priority process
     became runnable or the current process used up its time slice.  */
  long int ru_nivcsw;
};

int 
getrusage(int who, struct rusage *usage)
{
  gettimeofday(&usage->ru_utime, NULL);
  usage->ru_stime.tv_sec = 0;
  usage->ru_stime.tv_nsec = 0;
  return 0;
}

#define Undefined(name) void name() {report_unimplemented(#name);}
Undefined(__tcgetattr)
Undefined(acosh)
Undefined(atanh)
Undefined(asinh)

struct passwd {
  char    *pw_name;       /* user name */
  char    *pw_passwd;     /* user password */
  int   pw_uid;         /* user id */
  int   pw_gid;         /* group id */
  char    *pw_gecos;      /* real name */
  char    *pw_dir;        /* home directory */
  char    *pw_shell;      /* shell program */
};

struct passwd
fmh = {
  "byers",
  "no",
  4954,
  4023,
  "guess",
  "/home/byers",
  "sh"
};

struct passwd *
getpwuid(int uid)
{
  if (uid = fmh.pw_uid) {
    return &fmh;
  }
  return NULL;
}

int
getuid()
{
  char host[512];
  int uid, gid, ngids = 0, gidbuf[100];

  nfsAuthUnixGet(host, &uid, &gid, &ngids, gidbuf);
  return uid;
}




static char *
canonicalize (const char *name, char *resolved)
{
  char *rpath, *dest, *extra_buf = NULL;
  const char *start, *end, *rpath_limit;
  long int path_max;
  int num_links = 0;

  if (name == NULL)
    {
      /* As per Single Unix Specification V2 we must return an error if
	 either parameter is a null pointer.  We extend this to allow
	 the RESOLVED parameter be NULL in case the we are expected to
	 allocate the room for the return value.  */
      errnoSet (EINVAL);
      return NULL;
    }

  if (name[0] == '\0')
    {
      /* As per Single Unix Specification V2 we must return an error if
	 the name argument points to an empty string.  */
      errnoSet (ENOENT);
      return NULL;
    }

  path_max = PATH_MAX;

  rpath = resolved ? alloca (path_max) : malloc (path_max);
  rpath_limit = rpath + path_max;

  if (name[0] != '/')
    {
      if (!xgetcwd (rpath, path_max))
	goto error;
      dest = strchr (rpath, '\0');
    }
  else
    {
      rpath[0] = '/';
      dest = rpath + 1;
    }

  for (start = end = name; *start; start = end)
    {
      struct stat st;
      int n;

      /* Skip sequence of multiple path-separators.  */
      while (*start == '/')
	++start;

      /* Find end of path component.  */
      for (end = start; *end && *end != '/'; ++end)
	/* Nothing.  */;

      if (end - start == 0)
	break;
      else if (end - start == 1 && start[0] == '.')
	/* nothing */;
      else if (end - start == 2 && start[0] == '.' && start[1] == '.')
	{
	  /* Back up to previous component, ignore if at root already.  */
	  if (dest > rpath + 1)
	    while ((--dest)[-1] != '/');
	}
      else
	{
	  size_t new_size;

	  if (dest[-1] != '/')
	    *dest++ = '/';

	  if (dest + (end - start) >= rpath_limit)
	    {
	      ptrdiff_t dest_offset = dest - rpath;

	      if (resolved)
		{
		  errnoSet (ENAMETOOLONG);
		  goto error;
		}
	      new_size = rpath_limit - rpath;
	      if (end - start + 1 > path_max)
		new_size += end - start + 1;
	      else
		new_size += path_max;
	      rpath = realloc (rpath, new_size);
	      rpath_limit = rpath + new_size;
	      if (rpath == NULL)
		return NULL;

	      dest = rpath + dest_offset;
	    }

	  dest = memcpy (dest, start, end - start);
	  dest += (end-start);
	  *dest = '\0';

	  if (stat (rpath, &st) < 0)
	    goto error;

#ifdef VXWORKS_RESOLVES_LINKS
	  if (S_ISLNK (st.st_mode))
	    {
	      char *buf = __alloca (path_max);
	      size_t len;

	      if (++num_links > MAXSYMLINKS)
		{
		  errnoSet (ELOOP);
		  goto error;
		}

	      n = __readlink (rpath, buf, path_max);
	      if (n < 0)
		goto error;
	      buf[n] = '\0';

	      if (!extra_buf)
		extra_buf = __alloca (path_max);

	      len = strlen (end);
	      if ((long int) (n + len) >= path_max)
		{
		  errnoSet (ENAMETOOLONG);
		  goto error;
		}

	      /* Careful here, end may be a pointer into extra_buf... */
	      memmove (&extra_buf[n], end, len + 1);
	      name = end = memcpy (extra_buf, buf, n);

	      if (buf[0] == '/')
		dest = rpath + 1;	/* It's an absolute symlink */
	      else
		/* Back up to previous component, ignore if at root already: */
		if (dest > rpath + 1)
		  while ((--dest)[-1] != '/');
	    }
#endif
	}
    }
  if (dest > rpath + 1 && dest[-1] == '/')
    --dest;
  *dest = '\0';

  return resolved ? memcpy (resolved, rpath, dest - rpath + 1) : rpath;

error:
  if (resolved)
    strcpy (resolved, rpath);
  else
    free (rpath);
  return NULL;
}

/* We don't seem to have any way of reading links */
char *
realpath(char *pathname, char *resolvedname)
{
  if (resolvedname == NULL)
    {
      errnoSet (EINVAL);
      return NULL;
    }

  return canonicalize (pathname, resolvedname);
}
#endif
#endif

#ifdef LINUX
#include <netdb.h>
#include <sys/socket.h>
int
lookuphost(char *name)
{
  struct hostent *h = gethostbyname(name);
  if (h == NULL) {
    return (-1);
  }
  return *((int*) (h->h_addr));
}
#endif

#ifdef LINUX
struct dirent *
x_readdir(DIR *d)
{
  return readdir(d);
}
#endif

#ifdef VXWORKS
struct linux_dirent {
  int d_ino;
  int d_off;
  unsigned short d_reclen;
  unsigned char d_type;
  char d_name[256];
};

struct linux_dirent *
x_readdir(DIR *d)
{
  static struct linux_dirent ans;
  struct dirent *sys = readdir(d);

  if (sys == NULL) {
    return NULL;
  }
  ans.d_ino = 0;
  ans.d_off = 0;
  ans.d_type = 0;
  ans.d_reclen = 0;
  strcpy(ans.d_name, sys->d_name);
  return &ans;
}
#endif

#ifdef VXWORKS
int
lookuphost(char *name)
{
  int addr = inet_addr(name);
  if (addr == ERROR) {
    addr = hostGetByName(name);
  }
  return addr;
}
#endif

int
fd_setsize_bytes()
{
  return FD_SETSIZE/8;
}

void
do_fd_set(int fd, fd_set *fdsetp)
{
  FD_SET(fd, fdsetp);
}

void
do_fd_clr(int fd, fd_set *fdsetp)
{
  FD_CLR(fd, fdsetp);
}

int
do_fd_is_set(int fd, fd_set *fdsetp)
{
  return FD_ISSET(fd,fdsetp);
}

void
do_fd_zero(fd_set *fdsetp)
{
  FD_ZERO(fdsetp);
}
