/*
 * TCP socket library routines
 *
 * Copyright (c) 1998 by Erann Gat, all rights reserved
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifndef VXWORKS
#include <netdb.h>
#endif
#ifdef VXWORKS
#include <hostLib.h>
#endif

#include <sys/ioctl.h>
#include <unistd.h>
#include "sockets.h"

int open_client_socket(char *host, int port_num, int streamp)
{
  struct sockaddr_in server;
#ifndef VXWORKS
  struct hostent *hp;
#endif
#ifdef VXWORKS
  int remhost;
#endif
  int sock;

  /* Look up the server's IP address*/
#ifndef VXWORKS
  hp = gethostbyname(host);
  if (hp == NULL) {
    printf("Host not found.\n");
    return -1;
  }

#ifdef DEBUG
  {
    unsigned char *c = (char *)hp->h_addr;
    printf("Server IP address is %d.%d.%d.%d\n", c[0], c[1], c[2], c[3]);
  }
#endif

  /* Set up the sockaddr structure */
  bzero(&server, sizeof server);
  bcopy((char *)hp->h_addr, (char *)&server.sin_addr, hp->h_length);
  server.sin_family = AF_INET;
  server.sin_port = htons(port_num);
#endif

#ifdef VXWORKS
  remhost = hostGetByName(host);
  if (remhost == ERROR) {
    printf("Host not found.\n");
    return -1;
  }

  /* Set up the sockaddr structure */
  bzero(&server, sizeof server);
  bcopy((char *)&remhost, (char *)&server.sin_addr, sizeof(remhost));
  server.sin_family = AF_INET;
  server.sin_port = htons(port_num);


#endif
  
  /* create socket for the connection */
  if ((sock = socket(AF_INET,
		     streamp ? SOCK_STREAM : SOCK_DGRAM,
		     streamp ? IPPROTO_TCP : IPPROTO_UDP)) < 0) {
    perror("open_client_socket");
    return -1;
  }

  /* connect to the server */
  if (connect(sock, (struct sockaddr *)&server, sizeof server) < 0) {
    perror("open_client_socket");
    close(sock);
    return -1;
  }

  return sock;
}


int open_server_socket(int port_num)
{
  int sock;
  struct sockaddr_in server;
  struct linger linger_buff;
  int reuse = 1;
  int keepalive = 1;
  
  /* create socket */
  if ((sock = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0) {
    perror("open_server_socket");
    return -1;
  }
  
  /* set up internet address */
  bzero(&server,sizeof server);
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(port_num);
  
  /* set up socket options */
  
  /* allow reuse of local address when binding socket */
  if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&reuse,
		 sizeof reuse) < 0) {
    perror("open_server_socket");
    close(sock);
    return -1;
  }
  
  linger_buff.l_onoff = 0;
  linger_buff.l_linger = 0;
  
  /* dont wait for data trans when close is called */
  if (setsockopt(sock,SOL_SOCKET,SO_LINGER,(char *)&linger_buff,
		 sizeof linger_buff) < 0) {
    perror("open_server_socket");
    close(sock);
    return -1;
  }

  /* check periodically for connection and signal if lost */
  if (setsockopt(sock,SOL_SOCKET,SO_KEEPALIVE,(char *)&keepalive,
		 sizeof keepalive) < 0) {
    perror("open_server_socket");
    close(sock);
    return -1;
  }

  /* bind the address so clients can connect */
  if (bind(sock, (struct sockaddr *)&server, sizeof server) < 0) {
    perror("open_server_socket");
    close(sock);
    return -1;
  }
  
  /* listen for client to connect */
  if (listen(sock,5) < 0) {
    perror("open_server_socket");
    return -1;
  }
  
  return sock;
}



int accept_client_socket(int sock)
{
  int new_sock;
  struct sockaddr_in client;
  int client_length;
  
  /* zero out the socket address structure */
  bzero(&client,sizeof client);
  
  client_length = sizeof(client);
  if ((new_sock = accept(sock, (struct sockaddr *)&client, &client_length))
      < 0) {
    perror("accept_client_socket");
    return -1;
  }
    
  return new_sock;
}


int create_server_socket(int service)
{
  int server;
  
  server = open_server_socket(service);
  return accept_client_socket(server);
}


int open_tcp_client(char *server, int port)
{
  return open_client_socket(server, port, 1);
}

int open_udp_client(char *server, int port)
{
  return open_client_socket(server, port, 0);
}
