

/*	Change History (most recent first): */
/*  3 12/12/95 akh  gary's fix to tcallnfnslide */
/*  (do not edit before this line!!) */
/* Subprims for various flavors of function call, tail-recursion, etc. */

/* 06/19/96 bill Add some comments about stack overflow assumptions. */
/* ------------- MCL-PPC 3.9 */
/* 04/02/96 bill lexpr_entry initializes lisp_frame.savefn instead of lisp_frame.savelr twice */
/* 03/20/96 gb  .SPreset does (throw :toplevel $xstkover) */
/* 02/05/96 bill Gary's fix to lexpr_entry */
/* 01/29/96 gb  less hysteria in lexpr_entry */
/* 01/18/96 gb  add restorecontext, lexpr_entry */
/* 12/13/95  gb  udfcall->reset; callbuiltin, context stuff. */
/* 11/29/95  gb use jump_fname, jump_nfn macros */

	include(lisp.s)
	_beginfile
	.globl funcall
	.globl jmpsym


/* Note: all the tXXXslide functions MUST use a variable other than VSP while pushing to the slid location. */
/* This causes do_vsp_overflow in "lisp-exceptions.c" to allocate a new segment for the overflow, ensuring */
/* that it won't overwrite the stuff being slid. */
/* (They have to do so anyway, so that the VSP protects everything during the slide, but someone may */
/* need to remember the do_vsp_overflow assumption sometime). */

/* Tail-recursively call the (known symbol) in fname. */
	/* In the general case, we don't know if any args were */
	/* vpushed or not.  If so, we have to "slide" them down */
	/* to the base of the frame.  If not, we can just restore */
	/* vsp, lr, fn from the saved lisp frame on the control stack. */
_spentry(tcallsymgen)
	cmpwi cr0,nargs,nargregs<<fixnumshift
	ble cr0,._tcallsymvsp
	/* b tcallsymslide */

	/* Some args were vpushed.  Slide them down to the base of */
	/* the current frame, then do funcall. */
_spentry(tcallsymslide)
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz imm0,lisp_frame.savevsp(sp)
	discard_lisp_frame
	mtlr loc_pc
	/* can use nfn (= temp2) as a temporary */
	subi imm1,nargs,nargregs<<fixnumshift
	add imm1,imm1,vsp
1:	lwzu temp2,-4(imm1)
	cmpw cr0,imm1,vsp
	push temp2,imm0
	bne cr0,1b
	mr vsp,imm0
	b jmpsym

	/* No args were vpushed; recover saved context & call symbol */
_spentry(tcallsymvsp)
._tcallsymvsp:	
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz vsp,lisp_frame.savevsp(sp)
	discard_lisp_frame
	mtlr loc_pc
	/* b jmpsym */

_spentry(jmpsym)
	jump_fname

/* Tail-recursively call the function in nfn. */
	/* Pretty much the same as the tcallsym* cases above. */
_spentry(tcallnfngen)
	cmpwi cr0,nargs,nargregs<<fixnumshift
	ble cr0,.tcallnfnvsp
	/* b tcallnfnslide */

	/* Some args were vpushed.  Slide them down to the base of */
	/* the current frame, then do funcall. */
_spentry(tcallnfnslide)
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz imm0,lisp_frame.savevsp(sp)
	discard_lisp_frame
	mtlr loc_pc
	/* Since we have a known function, can use fname as a temporary. */
	subi imm1,nargs,nargregs<<fixnumshift
	add imm1,imm1,vsp
1:	lwzu fname,-4(imm1)
	cmpw cr0,imm1,vsp
	push fname,imm0
	bne cr0,1b
	mr vsp,imm0
	b jmpnfn

_spentry(tcallnfnvsp)
.tcallnfnvsp:
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz vsp,lisp_frame.savevsp(sp)
	discard_lisp_frame
	mtlr loc_pc
	/* b jmpnfn */

_spentry(jmpnfn)
	jump_nfn


/* Tail-recursively funcall temp0. */
	/* Pretty much the same as the tcallsym* cases above. */
_spentry(tfuncallgen)
	cmpwi cr0,nargs,nargregs<<fixnumshift
	ble cr0,.tfuncallvsp
	/* b tfuncallslide */

	/* Some args were vpushed.  Slide them down to the base of */
	/* the current frame, then do funcall. */
_spentry(tfuncallslide)
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz imm0,lisp_frame.savevsp(sp)
	discard_lisp_frame
	mtlr loc_pc
	/* can use nfn (= temp2) as a temporary */
	subi imm1,nargs,nargregs<<fixnumshift
	add imm1,imm1,vsp
1:	lwzu temp2,-4(imm1)
	cmpw cr0,imm1,vsp
	push temp2,imm0
	bne cr0,1b
	mr vsp,imm0
	b funcall

	/* No args were vpushed; recover saved context & do funcall */
_spentry(tfuncallvsp)
.tfuncallvsp:
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz fn,lisp_frame.savefn(sp)
	lwz vsp,lisp_frame.savevsp(sp)
	mtlr loc_pc
	discard_lisp_frame
	/* b funcall */

	/* Call temp0 if it's either a symbol or function */
_spentry(funcall)
	extract_lisptag imm0,temp0
	cmpwi imm0,tag_misc
	mr fname,temp0
	bne- 2f
	extract_subtag imm0,temp0
	cmpwi imm0,subtag_symbol
	cmpwi cr1,imm0,subtag_function
	bne- cr0,1f
	jump_fname
1:
	mr nfn,temp0
	bne- cr1,2f
	jump_nfn
2:	uuo_interr error_cant_call,temp0
	
_spentry(reset)
	.globl _SPthrow
	nop
/* rnil MUST be lisp_nil here.  When debugging, you */
/* should check that the low nibble of rnil (aka rtoc, */
/* aka r2) = #x5 or #xd.  You may be able to find */
/* the "real" value of lisp_nil from the pmcl-kernel */
/* debugger info file (pmcl-kernel.xcoff/pmcl-kernel.xsym); */
/* use some command named "show globals" or somesuch. */
	ref_nrs_value temp0,toplcatch
	li temp1,XSTKOVER
	vpush temp0
	vpush temp1
	set_nargs 1
	b _SPthrow
	
/* the value of the nilreg-relative symbol %builtin-functions% should be */
/* a vector of symbols.  Funcall the symbol indexed by imm0 (boxed) and */
/* return a single value. */

_spentry(callbuiltin0)
	set_nargs 0
	b callbuiltin

_spentry(callbuiltin1)
	set_nargs 1
	b callbuiltin

_spentry(callbuiltin2)
	set_nargs 2
	b callbuiltin

_spentry(callbuiltin3)
	set_nargs 3
_spentry(callbuiltin)
	ref_nrs_value fname,builtin_functions
	la imm0,misc_data_offset(imm0)
	lwzx fname,fname,imm0
	jump_fname

_spentry(popj)
	.globl .popj
.popj:	
	lwz loc_pc,lisp_frame.savelr(sp)
	mtlr loc_pc
	lwz vsp,lisp_frame.savevsp(sp)
	lwz fn,lisp_frame.savefn(sp)
	discard_lisp_frame
 .if 0
	ref_global imm0,current_cs
	lwz imm1,area.younger(imm0)
	cmpwi cr1,imm1,0
	lwz imm0,area.high(imm0)
	cmpwi imm0,0
	bnelr cr1
	lwz imm1,0(sp)
	beqlr
1:	cmplw cr0,imm1,imm0
	lwz imm2,0(imm1)
	cmplw cr1,imm1,imm2
	beqlr cr0
	crand 4,0,4
	mr imm3,imm1
	mr imm1,imm2
	blt+ cr1,1b
	.long 0
 .endif
	blr

_spentry(restorefullcontext)
	mflr loc_pc
	mtctr loc_pc
	lwz loc_pc,lisp_frame.savelr(sp)
	mtlr loc_pc
	lwz vsp,lisp_frame.savevsp(sp)
	lwz fn,lisp_frame.savefn(sp)
	discard_lisp_frame
 .if 0
	ref_global imm0,current_cs
	lwz imm1,area.younger(imm0)
	cmpwi cr1,imm1,0
	lwz imm0,area.high(imm0)
	cmpwi imm0,0
	bnelr cr1
	lwz imm1,0(sp)
	beqlr
1:	cmplw cr0,imm1,imm0
	lwz imm2,0(imm1)
	cmplw cr1,imm1,imm2
	beqlr cr0
	crand 4,0,4
	mr imm3,imm1
	mr imm1,imm2
	blt+ cr1,1b
	.long 0
 .endif
	bctr	

/* Like .SPrestorefullcontext, only the saved return address */
/* winds up in loc-pc instead of getting thrashed around ... */
_spentry(restorecontext)
	lwz loc_pc,lisp_frame.savelr(sp)
	lwz vsp,lisp_frame.savevsp(sp)
	lwz fn,lisp_frame.savefn(sp)
	discard_lisp_frame
 .if 0
	ref_global imm0,current_cs
	lwz imm1,area.younger(imm0)
	cmpwi cr1,imm1,0
	lwz imm0,area.high(imm0)
	cmpwi imm0,0
	bnelr cr1
	lwz imm1,0(sp)
	beqlr
1:	cmplw cr0,imm1,imm0
	lwz imm2,0(imm1)
	cmplw cr1,imm1,imm2
	beqlr cr0
	crand 4,0,4
	mr imm3,imm1
	mr imm1,imm2
	blt+ cr1,1b
	.long 0
 .endif
	blr
	
_spentry(savecontextvsp)
	li imm0,0
	/* b savecontext0 */

_spentry(savecontext0)
	create_lisp_frame
	add imm0,vsp,imm0
	stw fn,lisp_frame.savefn(sp)	
	stw loc_pc,lisp_frame.savelr(sp)
	stw imm0,lisp_frame.savevsp(sp)
	mr fn,nfn
	ref_global imm0,cs_overflow_limit
	twllt sp,imm0
 .if 0
	ref_global imm0,current_cs
	lwz imm1,area.younger(imm0)
	cmpwi cr1,imm1,0
	lwz imm0,area.high(imm0)
	cmpwi imm0,0
	bnelr cr1
	lwz imm1,0(sp)
	beqlr
1:	cmplw cr0,imm1,imm0
	lwz imm2,0(imm1)
	cmplw cr1,imm1,imm2
	beqlr cr0
	crand 4,0,4
	mr imm3,imm1
	mr imm1,imm2
	blt+ cr1,1b
	.long 0
 .endif
	blr

/* Nargs is valid; all arg regs, lexpr-count pushed by caller. */
/* imm0 = vsp to restore. */
/* Return all values returned by caller to its caller, hiding */
/* the variable-length arglist. */
/* If we can detect that the caller's caller didn't expect */
/* multiple values, then things are even simpler. */
_spentry(lexpr_entry)
	ref_global imm1,ret1val_addr
	cmpw cr0,imm1,loc_pc
	create_lisp_frame
	stw fn,lisp_frame.savefn(sp)
	stw loc_pc,lisp_frame.savelr(sp)
	stw imm0,lisp_frame.savevsp(sp)
	bne cr0,1f
	li loc_pc,fulltag_misc
	ref_global imm0,lexpr_return
	create_lisp_frame
	stw loc_pc,lisp_frame.savefn(sp)
	stw imm0,lisp_frame.savelr(sp)
	stw vsp,lisp_frame.savevsp(sp)
	mr loc_pc,imm1
	ref_global imm0,cs_overflow_limit
	twllt sp,imm0
	li fn,0
 .if 0
	ref_global imm0,current_cs
	lwz imm1,area.younger(imm0)
	cmpwi cr1,imm1,0
	lwz imm0,area.high(imm0)
	cmpwi imm0,0
	bnelr cr1
	lwz imm1,0(sp)
	beqlr
2:	cmplw cr0,imm1,imm0
	lwz imm2,0(imm1)
	cmplw cr1,imm1,imm2
	beqlr cr0
	crand 4,0,4
	mr imm3,imm1
	mr imm1,imm2
	blt+ cr1,2b
	.long 0
 .endif
	blr

/* The single-value case just needs to return to something that'll pop */
/* the variable-length frame off of the vstack. */
1:
	ref_global loc_pc,lexpr_return1v
	ref_global imm0,cs_overflow_limit
	twllt sp,imm0
	li fn,0
	blr

	
	
	_endfile

