
/* Stack-consing subprims (except for those associated with &rest args.) */

/* 06/22/96 bill  stkvcell0 adjusts vcell location for possible stack overflow. 
   06/19/96 bill  stkconsyz & stkvcellvsp reserve space on VSP before determining if address is odd. 
                  This is necessary since the oddness could change if pushing allocates a new stack segment. 
                  stack_misc_alloc_init explicitly restores vsp before exiting in case 
                  the init code allocated a new stack frame. 
   06/14/96  bill makestacklist, stkgvector, stkconslist_star, mkstackv, stack_misc_alloc 
                  initialize initptr after, not before, creating a new tsp frame. 
                  remove the "half-hearted stack probe"s from stkconslist_star & mkstackv. 
                  makestackblock & makestackblock0 limit stack consing to 4K bytes. 
                  They heap cons a gcable macptr for larger requests. 
                  makestackblock0 works correctly if block length is 0. 
   -------------  MCL-PPC 3.9 
   02/06/96  bill TSP overflow checking for makestackblock, makestackblock0, makestacklist 
   11/29/95  gb   stack-consed vectors 
   10/28/95  gb   push value_cell_header in stkvcell0 */

	include(lisp.s)
	_beginfile
	.globl init_list_from_stack
	.globl _SPmisc_alloc
	
/* Make a cons cell on the vstack.  Always push 3 words, 'cause we're  
   not sure how the vstack will be aligned. */
_spentry(stkconsyz)
	vpush rnil
	vpush rnil
	vpush rnil
	andi. imm0,vsp,1<<2	/* (oddp vsp ?) */
	beq cr0,1f
	stw arg_y,8(vsp)	/* car */
	stw arg_z,4(vsp)	/* cdr */
	la arg_z,fulltag_cons+4(vsp)
	blr
1:	stw arg_y,4(vsp)	/* car, again */
	stw arg_z,0(vsp)
	la arg_z,fulltag_cons(vsp)
	blr

/* Make a stack-consed value cell.  Much like the case of stack-allocating a cons 
   cell.  Imm0 points to the closed-over value (already vpushed).  Replace 
   that locative with the vcell. */
_spentry(stkvcellvsp)
	mr imm0,vsp
_spentry(stkvcell0)
	sub imm1,imm0,vsp	/* imm1 = delta from vsp to value cell loc */
	vpush rnil
	vpush rnil
	vpush rnil
	addi imm1,imm1,12
	add imm0,vsp,imm1	/* in case stack overflowed */
	andi. imm1,vsp,1<<2	/* (oddp vsp) ? */
	li imm1,value_cell_header
	lwz arg_z,0(imm0)
	beq cr0,1f
	stw arg_z,8(vsp)
	stw imm1,4(vsp)
	la arg_z,fulltag_misc+4(vsp)
	stw arg_z,0(imm0)
	blr
1:	stw arg_z,4(vsp)
	stw imm1,0(vsp)
	la arg_z,fulltag_misc(vsp)
	stw arg_z,0(imm0)
	blr

/* Make a "raw" area on the temp stack, stack-cons a macptr to point to it, 
   and return the macptr.  Size (in bytes, boxed) is in arg_z on entry; macptr in 
   arg_z on exit. 
   It would be nice to cons in the Mac heap if there's not room on 
   the tstack. This code will handle a new tstack segment being added. */
_spentry(makestackblock)
	unbox_fixnum imm0,arg_z
	la imm0,8+macptr.size+7(imm0)
	clrrwi imm0,imm0,3
	cmpwi cr0,imm0,4096
	neg imm0,imm0
	bge cr0,1f

	stwux tsp,tsp,imm0
	stw tsp,4(tsp)		/* "raw" block. */
	li imm0,macptr_header
	la imm1,8+macptr.size(tsp)
	la arg_z,8+fulltag_misc(tsp)
	stw imm0,macptr.header(arg_z)
	stw imm1,macptr.address(arg_z)   /* makestackblock0 expects the address to be in imm1 */
	blr

/* Too big. Heap cons a gcable macptr */
1:
	stwu tsp,-8(tsp)
	stw tsp,4(tsp)          /* "raw" block to make the compiler happy */
	set_nargs 1
	la fname,nrs.new_gcable_ptr(rnil)
	jump_fname
	

/* As above, only set the block's contents to 0. */
_spentry(makestackblock0)
	mr arg_y,arg_z                         /* save block size */
	unbox_fixnum imm0,arg_z
	la imm0,8+macptr.size+7(imm0)
	clrrwi imm0,imm0,3
	cmpwi cr0,imm0,4096
	neg imm0,imm0
	bge cr0,3f

	stwux tsp,tsp,imm0
	stw tsp,4(tsp)                         /* "raw" block. */
	li imm0,macptr_header
	la imm1,8+macptr.size(tsp)
	la arg_z,8+fulltag_misc(tsp)
	stw imm0,macptr.header(arg_z)
	stw imm1,macptr.address(arg_z)         /* makestackblock0 expects the address to be in imm1 */

	unbox_fixnum imm0,arg_y
	la imm0,7(imm0)
	clrrwi imm0,imm0,3
	add imm2,imm1,imm0                     /* end of the line */
	cmpw cr0,imm1,imm2
	la imm2,-8(imm2)
	b 2f
1:
	cmpw cr0,imm1,imm2
	stw rzero,0(imm1)
	la imm1,8(imm1)
	stw rzero,-4(imm1)
2:	bne cr0,1b
	blr
	
/* Too big. Heap cons a gcable macptr */
3:
	stwu tsp,-8(tsp)
	stw tsp,4(tsp)          /* "raw" block to make the compiler happy */
	
	la arg_z,nrs.tsym(rnil) /* clear-p arg to %new-gcable-ptr */
	set_nargs 2
	la fname,nrs.new_gcable_ptr(rnil)
	jump_fname

/* Make a list of length arg_y (boxed), initial-element arg_z (boxed) on 
   the tstack.  Return the list in arg_z. */
_spentry(makestacklist)
	add imm0,arg_y,arg_y
	cmpwi cr1,imm0,4096-8
	addi imm0,imm0,8
	neg imm0,imm0
	bge cr1,3f
	mr imm1,arg_y
	cmpwi cr1,imm1,0
	mr arg_y,arg_z
	mr arg_z,rnil
	mr initptr,tsp			/* uninterruptable */
	stwux tsp,tsp,imm0
	stw rzero,4(tsp)
	sub initptr,tsp,imm0		/* correct for stack overflow */
	la initptr,-8+tag_list(initptr)
	b 2f
1:	subi imm1,imm1,fixnum1
	cmpwi cr1,imm1,0
	rplacd initptr,arg_z
	rplaca initptr,arg_y
	mr arg_z,initptr
	subi initptr,initptr,cons.size
2:	bne cr1,1b
	mr initptr,freeptr
	blr

3:
        cmpwi cr1,arg_y,0
	stwu tsp,-8(tsp)                       /* make the compiler happy */
	stw rzero,4(tsp)
	mr imm1,arg_y                          /* count */
	mr arg_y,arg_z                         /* initial value */
	mr arg_z,rnil                          /* result */
        b 5f
4:
	subi imm1,imm1,fixnum1
	cmpwi cr1,imm1,0
	stwu rzero,cons.size(freeptr)
	la initptr,tag_list(initptr)
	rplaca initptr,arg_y
	rplacd initptr,arg_z
	mr arg_z,initptr
	mr initptr,freeptr
5:
	bne cr1,4b
	blr


/* subtype (boxed) vpushed before initial values. (Had better be a 
	node header subtag.) Nargs set to count of things vpushed. */
_spentry(stkgvector)
	la imm0,-4(nargs)
	cmpwi cr0,imm0,0
	add imm1,vsp,nargs
	lwzu temp0,-4(imm1)
	slwi imm2,imm0,num_subtag_bits-fixnumshift
	rlwimi imm2,temp0,32-fixnumshift,32-num_subtag_bits,31
	la imm0,(8+4+7)(imm0)
	clrrwi imm0,imm0,3
	neg imm0,imm0
	mr initptr,tsp			/* uninterruptable */
	stwux tsp,tsp,imm0
	sub initptr,tsp,imm0		/* correct for stack overflow */
	stw rzero,-4(initptr)
	stw rzero,4(tsp)
	la arg_z,8+fulltag_misc(tsp)
	la initptr,misc_header_offset(arg_z)
	stw imm2,0(initptr)
	li imm0,fixnum1
	b 2f
1:	addi imm0,imm0,fixnum1
	cmpw cr0,imm0,nargs
	lwzu temp0,-4(imm1)
	stwu temp0,4(initptr)
2:	bne cr0,1b
	mr initptr,freeptr
	add vsp,vsp,nargs
	blr


/* We always have to create a tsp frame (even if nargs is 0), so the compiler 
   doesn't get confused. */
_spentry(stkconslist)
	mr arg_z,rnil
/* do list*: last arg in arg_z, all others vpushed, 
	nargs set to #args vpushed. */
_spentry(stkconslist_star)
	/*stw rzero,-4096(tsp)		; half-hearted stack probe. */
	cmpwi nargs,0
	add imm1,nargs,nargs
	addi imm1,imm1,8
	neg imm1,imm1
	mr initptr,tsp			/* uninterruptable */
	stwux tsp,tsp,imm1
	sub initptr,tsp,imm1		/* correct for stack overflow */
	la initptr,-8+fulltag_cons(initptr)
	stw rzero,4(tsp)
	b init_list_from_stack

/* Make a stack-consed simple-vector out of the NARGS objects 
	on top of the vstack; it in arg_z. */
_spentry(mkstackv)
	/*stw rzero,-4096(tsp) */
	cmpwi nargs,0
	neg imm1,nargs
	la imm1,-(4+8)(imm1)
	bitclr imm1,imm1,2
	mr initptr,tsp			/* uninterruptable */
	stwux tsp,tsp,imm1
	sub initptr,tsp,imm1		/* correct for stack overflow */
	stw rzero,4(tsp)
	stw rzero,-4(initptr)
	slwi imm0,nargs,num_subtag_bits-fixnumshift
	ori imm0,imm0,subtag_simple_vector
	stw imm0,8(tsp)
	la arg_z,8+fulltag_misc(tsp)
	beq- 2f
	la imm0,8+4(tsp)
	add initptr,imm0,nargs
1:
	la nargs,-4(nargs)
	cmpwi nargs,0
	lwz temp1,0(vsp)
	la vsp,4(vsp)
	stwu temp1,-4(initptr)
	bne 1b
2:	mr initptr,freeptr
	blr

/* Allocate a miscobj on the temp stack.  (Push a frame on the tsp and 
   heap-cons the object if there's no room on the tstack.) */

_spentry(stack_misc_alloc)
	rlwinm. imm2,arg_y,32-fixnumshift,0,(8+fixnumshift)-1
	unbox_fixnum imm0,arg_z
	extract_fulltag imm1,imm0
	bne- cr0,9f
	cmpwi cr0,imm1,fulltag_nodeheader
	mr imm3,imm0
	cmplwi cr1,imm0,max_32_bit_ivector_subtag
	rlwimi imm0,arg_y,num_subtag_bits-fixnum_shift,0,31-num_subtag_bits	/* imm0 now = header */
	mr imm2,arg_y
	beq cr0,1f	/* do probe if node object 
			   (fixnum element count = byte count). */
	cmplwi cr0,imm3,max_16_bit_ivector_subtag
	bng cr1,1f	/* do probe if 32-bit imm object */
	cmplwi cr1,imm3,max_8_bit_ivector_subtag
	srwi imm2,imm2,1
	bgt cr0,3f
	bgt cr1,1f
	srwi imm2,imm2,1
/* imm2 now = byte count.  Add 4 for header, 7 to align, then 
	clear low three bits. */
1:
	addi imm2,imm2,4+7
	clrrwi imm2,imm2,3
	la imm3,8(imm2)		/* account for tsp header */
	cmpwi cr0,imm3,0x1000	/* more than 4k ? */
	bgt- cr0,0f
	neg imm3,imm3
	mr initptr,tsp		/* no interrupts */
	stwux tsp,tsp,imm3
	sub initptr,tsp,imm3	/* correct for stack overflow */
	stw rzero,4(tsp)
/* Even if we're going to initialize this to some non-zero value in a 
	second, we initialize the vector's contents to 0 now, so that
	the period in which we're  gc-safe is as short as possible. 
	Note that the loop can be written to assume that we've always 
	got at least doubleword to initialize (because of the vector header.) 
*/
2:
	subi imm2,imm2,8
	cmpwi cr0,imm2,0
	subi initptr,initptr,8
	stw rzero,4(initptr)
	stw rzero,0(initptr)
	bne cr0,2b
/* Slap the header on the vector, then fix initptr & return. */
	la arg_z,fulltag_misc(initptr)
	mr initptr,freeptr
	stw imm0,misc_header_offset(arg_z)
	blr
9: 



/* Too large to safely fit on tstack.  Heap-cons the vector, but make 
   sure that there's an empty tsp frame to keep the compiler happy. */
0:	stwu tsp,-8(tsp)
	stw rzero,4(tsp)
	b _SPmisc_alloc
3:
	cmplwi imm3,subtag_double_float_vector
	slwi imm2,arg_y,1
	beq 1b
	addi imm2,arg_y,7<<fixnumshift
	srwi imm2,imm2,fixnumshift+3
	b 1b

/* As in stack_misc_alloc above, only with a non-default initial-value. */

_spentry(stack_misc_alloc_init)
	mflr loc_pc
	create_lisp_frame
	stw fn,lisp_frame.savefn(sp)
	stw loc_pc,lisp_frame.savelr(sp)
	stw vsp,lisp_frame.savevsp(sp)
	li fn,fulltag_misc
	mr temp0,arg_z		/* initval */
	mr arg_z,arg_y		/* subtag */
	mr arg_y,arg_x		/* element-count */
	bl stack_misc_alloc
	lwz loc_pc,lisp_frame.savelr(sp)
	mtlr loc_pc
	lwz fn,lisp_frame.savefn(sp)
	lwz vsp,lisp_frame.savevsp(sp) /* vsp may have changed due to overflowing a stack segment */
	discard_lisp_frame
	la fname,nrs.init_misc(rnil)
	set_nargs 2
	mr arg_y,temp0
	jump_fname


	_endfile


