
/* 06/10/96 gb  uuo_fixnum_overflow */
/* --- 3.9 release */

/* A uuo looks like: */
/*  0      5 6                  15 16   20 21          27 28  31 */
/* +--------+-----------------------------+--------------+------+ */
/* |   0    |XXXXXXXXXXXXXXXXXXXX |  RB   |  <minor op>  |  11  | */
/* +--------+-----------------------------+--------------+------+ */
/* */
/* e.g., the major opcode (bits 0-5) is 0, the low 4 bits (bits 28-31) */
/* have the value "11" decimal (that's tagged as an immediate as far */
/* as lisp is concerned, a 7-bit opcode in bits 21-27, and the format */
/* of bits 6-20 depend on the value of the minor opcode, though typically */
/* bits 16-20 are used to specify a register value between 0 and 31. */
/* */
/* There are a few cases where bits 6-15 are also used to denote registers */
/* (RT and RA, as in an X-form PPC instruction), some where bits 6-10 are */
/* to be interpreted as a constant (error number or type code), and some */
/* where bits 6-15 do so. */
/* */
/* Since C code is typically more interested in disassembling UUOs, the */
/* full list of UUOs is in "uuo.h".  This file contains macros for creating */
/* them. */
/* */
/* Of course, there -is- no such file as "uuo.h".  That's a stale comment. */
/* For all anyone knows, so is this one. */

UUO_TAG	= 11
UUU_MINOR_SHIFT = 4
UUO_RB_SHIFT = 11
UUO_RA_SHIFT = 16
UUO_RT_SHIFT = 21

	.macro rt_ra_uuo minorop,rt,ra
	.long (UUO_TAG|((\minorop)<<UUU_MINOR_SHIFT)|((\ra)<<UUO_RA_SHIFT)|((\rt)<<UUO_RT_SHIFT))
	.endm

	.macro rt_ra_rb_uuo minorop,rt,ra,rb
	.long (UUO_TAG|((\minorop)<<UUU_MINOR_SHIFT)|((\ra)<<UUO_RA_SHIFT)|((\rb)<<UUO_RB_SHIFT)|((\rt)<<UUO_RT_SHIFT))
	.endm
	
	.macro errnum_rb_uuo minorop,errnum,rb
	.long (UUO_TAG|((\minorop)<<UUU_MINOR_SHIFT)|((\errnum)<<UUO_RA_SHIFT)|((\rb)<<UUO_RB_SHIFT))
	.endm
	
	.macro errnum_ra_rb_uuo minorop,errnum,ra,rb
	.long (UUO_TAG|((\minorop)<<UUU_MINOR_SHIFT)|((\ra)<<UUO_RA_SHIFT)|((\rb)<<UUO_RB_SHIFT)|((\errnum)<<UUO_RT_SHIFT))
	.endm
	
	
	.macro uuo_box_signed dest,src
	rt_ra_uuo 1,\dest,\src
	.endm
	
	.macro uuo_box_unsigned dest,src
	rt_ra_uuo 2,\dest,\src
	.endm

	.macro uuo_fixnum_overflow dest,src
	rt_ra_uuo 21,\dest,\src
	.endm


	
/* Signal an internal error - type error or whatever - with error  */
/* number (0-1023) and "register" argument. */

	.macro uuo_interr errnum,rb
	errnum_rb_uuo 11,\errnum,\rb
	.endm
	
/* As above, but make the error continuable.  (A branch presumably */
/* follows the UUO opcode.) */

	.macro uuo_intcerr errnum,rb
	errnum_rb_uuo 12,\errnum,\rb
	.endm

/* Signal an error with a much smaller error number (0-31) and */
/* two "register" fields. */

	.macro uuo_interr2 errnum,ra,rb
	errnum_ra_rb_uuo 13,\errnum,\ra,\rb
	.endm
	
/* Continuably .... */

	.macro uuo_intcerr2 errnum,ra,rb
	errnum_ra_rb_uuo 14,\errnum,\ra,\rb
	.endm

/* Ask C to allocate a "large" (> 32K bytes) object.  Boxed element count in */
/* ra, boxed subtag in rb, boxed result goes in rt.  May GC or fail. */

	.macro uuo_xalloc rt,ra,rb
	rt_ra_rb_uuo 24,\rt,\ra,\rb
	.endm
	

