;;;-*- Mode: Lisp; Package: CCL -*-
(in-package "CCL")

(eval-when (:compile-toplevel :execute)
  (require "FASLENV" "ccl:xdump;faslenv")
  (require "PPC-LAP"))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (require "XFASLOAD" "ccl:xdump;xfasload"))

; This is used to pass the value of lisp_nil into the
; kernel.  The data section will look like:
;  dc.l (arch::make-vheader (+ 1024 3) arch::subtag-u32-vector)
;  dc.l import_ref_to "set_nil_and_start"
;  dc.l code_address, lisp_nil
;  ds.l 1024  ; lisp_globals
;   ds.b fulltag_nil
; lisp_nil:      ; our TOC anchor
; ...
; which explains what (- (+ arch::fulltag-nil 4096 12)) is all about.
(defparameter *ppc-application-entry-function-code*
  (uvref (%define-ppc-lap-function nil '((lwz imm1 (- (+ arch::fulltag-nil 4096 12)) rnil)
                                         (lwz imm2 0 imm1)
                                         (mr imm0 rnil)
                                         (mtctr imm2)
                                         (lwz rnil 4 imm1)
                                         (bctr)))
         0))

(defparameter *ppc-macro-apply-code*
  (#+ppc-target normalize-code-vector
   #-ppc-target identity
  (uvref
   (%define-ppc-lap-function  ()
                              '((li arg_y '#.$xnotfun)
                                (lwz arg_z (+ arch::misc-data-offset (ash 1 arch::word-shift)) nfn)
                                (set-nargs 2)
                                (ba .SPksignalerr)))
   0
   )))



(defparameter *ppc-closure-trampoline-code*
  (#+ppc-target normalize-code-vector
   #-ppc-target identity
  (uvref
   (%define-ppc-lap-function  ()
                              '((ba .SPcall-closure)))
   0
   )))




; For now, do this with a UUO so that the kernel can catch it.
(defparameter *ppc-udf-code*
  (#+ppc-target normalize-code-vector
   #-ppc-target identity
   (uvref (%define-ppc-lap-function nil '((uuo_interr #.arch::error-udf-call 0))) 0)))





(defparameter *ppc-excised-code*
  (#+ppc-target normalize-code-vector
   #-ppc-target identity
   (uvref (%define-ppc-lap-function nil '((uuo_interr #.arch::error-excised-function-call 0))) 0)))

; Slap a code-vector header and a NOP on the front of the subprims code.
; This assumes that the subprims (and the extra 8 bytes) will fit; in any
; reasonable world, it will.
(defun xppc-prepend-subprims-code (space)
  (let* ((data-vector (xload-space-data space))
         (raw-fullwords (* 256 4))
         (raw-size (* raw-fullwords 4)))
    (macrolet ((ppc-lap-word (instruction-form)
                   (uvref (uvref (compile nil
                                          `(lambda (&lap 0)
                                             (ppc-lap-function () ((?? 0))
                                                               ,instruction-form)))                                 
                                 0) 0)))
        ;; Create a jump table with 256 (= maxsubp) entries.  Each entry
        ;; branches to a 12-byte sequence, so the branch displacement increases by 8 each time
        (do* ((disp (* 256 4) (+ disp 8))
              (i 0 (1+ i))
              (offset 8 (+ offset 4)))
             ((= i 256))
          (setf (u32-ref data-vector offset) (logior (ash 18 26) disp)))         ; (b *+disp)
        (do* ((i 0 (1+ i))
              (offset (+ 8 (* 256 4)) (+ offset 12))
              (instr0 (ppc-lap-word (lwz imm4 (+ (- arch::fulltag-nil) (* -512 4)) rnil)) (+ instr0 4)))
             ((= i 256))
          (setf (u32-ref data-vector offset) instr0)
          (setf (u32-ref data-vector (+ offset 4)) (ppc-lap-word (mtctr imm4)))
          (setf (u32-ref data-vector (+ offset 8)) (ppc-lap-word (bctr)))))
    (setf (u32-ref data-vector 4) #x60000000)     ; a NOP
    (when (oddp raw-fullwords)
      (setf (u32-ref data-vector (+ 8 raw-size)) #x60000000)
      (incf raw-fullwords))
    (setf (u32-ref data-vector 0) (arch::make-vheader (1+ raw-fullwords) arch::subtag-code-vector))
    (setf (xload-space-lowptr space) (xload-dword-align (+ 8 raw-size) nil))))


               
; We're assuming here that the subprims jump table starts at byte-offset 8
; in the same data vector we're in.
; We reference the jump table entry, which isn't really necessary but
; (a) is a little easier and (b) makes incorporating debugging info
; for the subprims code a little easier.
(defun xppc-relativize-subprim-calls (data byte-offset nfullwords)
  (declare (fixnum byte-offset nfullwords))
  (do* ((i 0 (1+ i))
        (b byte-offset (+ b 4)))
       ((= i nfullwords))
    (declare (fixnum i b))
    (let* ((w (u32-ref data b))
           (major (ldb (byte 6 26) w)))
      (declare (fixnum major))
      (if (eql w 0) (return))
      (when (and (= major 18)           ; "b" opcode
                 (logbitp 1 w))         ; "aa" bit set
        (let* ((target (+ 8 (the fixnum (logand w (logand (1- (ash 1 26)) (lognot 3)))))))
          (declare (fixnum target))
          (setf (u32-ref data b)
                (dpb 18 (byte 6 26) (logior (the fixnum (logand w 1)) ; preserve "lk" bit
                                            (the fixnum (- target b))))))))))

(defparameter *ppc-xload-backend*
  (make-backend-xload-info
   :name :PPC
   :application-entry-code *ppc-application-entry-function-code*
   :macro-apply-code *ppc-macro-apply-code*
   :closure-trampoline-code *ppc-closure-trampoline-code*
   :udf-code *ppc-udf-code*
   :excised-code *ppc-excised-code*
   :default-image-name "ccl:ccl;ppc-boot"
   :default-startup-file-name "level-1.pfsl"
   :relativize-subprims-hook 'xppc-relativize-subprim-calls
   :prepend-subprims-hook 'xppc-prepend-subprims-code
   :subdir "ccl:level-0;PPC;"
   :compiler-target-name :ppc
))

(add-xload-backend *ppc-xload-backend*)

#+ppc-target
(progn
(setq *xload-default-backend* *ppc-xload-backend*)

(defun Xcompile-directory (dir &optional force)
  (target-xcompile-directory :ppc dir  force))

(defun Xcompile-level-0 (&optional force)
  (target-xcompile-level-0 :PPC force))

(defun xload-level-0 (&optional (recompile t))
  (target-xload-level-0 :ppc recompile))

(defun xload-bootstrap (&optional (recompile t))
  (let* ((bootstrap-backend (copy-backend-xload-info *ppc-xload-backend*)))
    (setf (backend-xload-info-name bootstrap-backend) :ppc-bootstrap
	  (backend-xload-info-default-image-name bootstrap-backend) "ccl:ccl;runtime-boot"
	  (backend-xload-info-default-startup-file-name bootstrap-backend) "runtime.pfsl")
    (add-xload-backend bootstrap-backend)
    (target-xload-level-0 :ppc-bootstrap recompile)))

)
